@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -27,6 +28,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <unistd.h>
 #include <msgpack.h>
 
@@ -34,7 +36,8 @@
 #include "kube_meta.h"
 #include "kube_property.h"
 
-static int file_to_buffer(char *path, char **out_buf, size_t *out_size)
+static int file_to_buffer(const char *path,
+                          char **out_buf, size_t *out_size)
 {
     int ret;
     char *buf;
@@ -136,10 +139,12 @@ static int get_local_pod_info(struct flb_kube *ctx)
 
 /* Gather metadata from API Server */
 static int get_api_server_info(struct flb_kube *ctx,
-                               char *namespace, char *podname,
+                               const char *namespace, const char *podname,
                                char **out_buf, size_t *out_size)
 {
     int ret;
+    int packed = -1;
+    int root_type;
     size_t b_sent;
     char uri[1024];
     char *buf;
@@ -147,60 +152,98 @@ static int get_api_server_info(struct flb_kube *ctx,
     struct flb_http_client *c;
     struct flb_upstream_conn *u_conn;
 
-    if (!ctx->upstream) {
-        return -1;
-    }
+    /*
+     * If a file exists called namespace-podname.meta, load it and use it.
+     * If not, fall back to API. This is primarily for diagnostic purposes,
+     * e.g. debugging new parsers.
+     */
+    if (ctx->meta_preload_cache_dir && namespace && podname) {
+        int fd;
+        char *payload = NULL;
+        size_t payload_size = 0;
+        struct stat sb;
 
-    u_conn = flb_upstream_conn_get(ctx->upstream);
-    if (!u_conn) {
-        flb_error("[filter_kube] upstream connection error");
-        return -1;
-    }
-
-    ret = snprintf(uri, sizeof(uri) - 1,
-                   FLB_KUBE_API_FMT,
-                   namespace, podname);
-    if (ret == -1) {
-        flb_upstream_conn_release(u_conn);
-        return -1;
-    }
-
-    /* Compose HTTP Client request */
-    c = flb_http_client(u_conn, FLB_HTTP_GET,
-                        uri,
-                        NULL, 0, NULL, 0, NULL, 0);
-    flb_http_buffer_size(c, ctx->buffer_size);
-
-    flb_http_add_header(c, "User-Agent", 10, "Fluent-Bit", 10);
-    flb_http_add_header(c, "Connection", 10, "close", 5);
-    if (ctx->auth_len > 0) {
-        flb_http_add_header(c, "Authorization", 13, ctx->auth, ctx->auth_len);
-    }
-
-    /* Perform request */
-    ret = flb_http_do(c, &b_sent);
-    flb_debug("[filter_kube] API Server (ns=%s, pod=%s) http_do=%i, HTTP Status: %i",
-              namespace, podname, ret, c->resp.status);
-
-    if (ret != 0 || c->resp.status != 200) {
-        if (c->resp.payload_size > 0) {
-            flb_debug("[filter_kube] API Server response\n%s",
-                      c->resp.payload);
+        ret = snprintf(uri, sizeof(uri) - 1, "%s/%s-%s.meta", ctx->meta_preload_cache_dir, namespace, podname);
+        if (ret > 0) {
+            fd = open(uri, O_RDONLY, 0);
+            if (fd > 0) {
+                if (fstat(fd, &sb) == 0) {
+                    payload = flb_malloc(sb.st_size);
+                    if (payload) {
+                        ret = read(fd, payload, sb.st_size);
+                        if (ret == sb.st_size) {
+                            payload_size = ret;
+                        }
+                    }
+                }
+                close(fd);
+            }
         }
+        if (payload_size) {
+            packed = flb_pack_json(payload, payload_size,
+                                   &buf, &size, &root_type);
+        }
+
+        if (payload) {
+            flb_free(payload);
+        }
+    }
+
+    if (packed == -1) {
+        if (!ctx->upstream) {
+            return -1;
+        }
+
+        u_conn = flb_upstream_conn_get(ctx->upstream);
+        if (!u_conn) {
+            flb_error("[filter_kube] upstream connection error");
+            return -1;
+        }
+
+        ret = snprintf(uri, sizeof(uri) - 1,
+                       FLB_KUBE_API_FMT,
+                       namespace, podname);
+        if (ret == -1) {
+            flb_upstream_conn_release(u_conn);
+            return -1;
+        }
+
+        /* Compose HTTP Client request */
+        c = flb_http_client(u_conn, FLB_HTTP_GET,
+                            uri,
+                            NULL, 0, NULL, 0, NULL, 0);
+        flb_http_buffer_size(c, ctx->buffer_size);
+
+        flb_http_add_header(c, "User-Agent", 10, "Fluent-Bit", 10);
+        flb_http_add_header(c, "Connection", 10, "close", 5);
+        if (ctx->auth_len > 0) {
+            flb_http_add_header(c, "Authorization", 13, ctx->auth, ctx->auth_len);
+        }
+
+        /* Perform request */
+        ret = flb_http_do(c, &b_sent);
+        flb_debug("[filter_kube] API Server (ns=%s, pod=%s) http_do=%i, HTTP Status: %i",
+                  namespace, podname, ret, c->resp.status);
+
+        if (ret != 0 || c->resp.status != 200) {
+            if (c->resp.payload_size > 0) {
+                flb_debug("[filter_kube] API Server response\n%s",
+                          c->resp.payload);
+            }
+            flb_http_client_destroy(c);
+            flb_upstream_conn_release(u_conn);
+            return -1;
+        }
+        packed = flb_pack_json(c->resp.payload, c->resp.payload_size,
+                               &buf, &size, &root_type);
+
+        /* release resources */
         flb_http_client_destroy(c);
         flb_upstream_conn_release(u_conn);
-        return -1;
     }
 
-    ret = flb_pack_json(c->resp.payload, c->resp.payload_size,
-                        &buf, &size);
-
-    /* release resources */
-    flb_http_client_destroy(c);
-    flb_upstream_conn_release(u_conn);
-
     /* validate pack */
-    if (ret == -1) {
+    if (packed == -1) {
         return -1;
     }
 
@@ -210,37 +253,37 @@ static int get_api_server_info(struct flb_kube *ctx,
     return 0;
 }
 
-static void cb_results(unsigned char *name, unsigned char *value,
+static void cb_results(const char *name, const char *value,
                        size_t vlen, void *data)
 {
     struct flb_kube_meta *meta = data;
 
-    if (meta->podname == NULL && strcmp((char *) name, "pod_name") == 0) {
-        meta->podname = flb_strndup((char *) value, vlen);
+    if (meta->podname == NULL && strcmp(name, "pod_name") == 0) {
+        meta->podname = flb_strndup(value, vlen);
         meta->podname_len = vlen;
         meta->fields++;
     }
     else if (meta->namespace == NULL &&
-             strcmp((char *) name, "namespace_name") == 0) {
-        meta->namespace = flb_strndup((char *) value, vlen);
+             strcmp(name, "namespace_name") == 0) {
+        meta->namespace = flb_strndup(value, vlen);
         meta->namespace_len = vlen;
         meta->fields++;
     }
     else if (meta->container_name == NULL &&
-             strcmp((char *) name, "container_name") == 0) {
-        meta->container_name = flb_strndup((char *) value, vlen);
+             strcmp(name, "container_name") == 0) {
+        meta->container_name = flb_strndup(value, vlen);
         meta->container_name_len = vlen;
         meta->skip++;
     }
     else if (meta->docker_id == NULL &&
-             strcmp((char *) name, "docker_id") == 0) {
-        meta->docker_id = flb_strndup((char *) value, vlen);
+             strcmp(name, "docker_id") == 0) {
+        meta->docker_id = flb_strndup(value, vlen);
         meta->docker_id_len = vlen;
         meta->skip++;
     }
     else if (meta->container_hash == NULL &&
-             strcmp((char *) name, "container_hash") == 0) {
-        meta->container_hash = flb_strndup((char *) value, vlen);
+             strcmp(name, "container_hash") == 0) {
+        meta->container_hash = flb_strndup(value, vlen);
         meta->container_hash_len = vlen;
         meta->skip++;
     }
@@ -248,8 +291,100 @@ static void cb_results(unsigned char *name, unsigned char *value,
     return;
 }
 
+/*
+ * As per Kubernetes Pod spec,
+ * https://kubernetes.io/docs/concepts/workloads/pods/pod/, we look
+ * for status.containerStatuses.{containerID, imageID} where
+ * status.containerStatus.name == our container name
+ * status:
+ *   ...
+ *   containerStatuses:
+ *   - containerID: XXX
+ *     image: YYY
+ *     imageID: ZZZ
+ *     ...
+ *     name: nginx-ingress-microk8s
+*/
+#define CONTAINER_STATUSES "containerStatuses"
+static void extract_container_hash(struct flb_kube_meta *meta,
+                                   msgpack_object status)
+{
+    int i;
+    msgpack_object k, v;
+    int docker_id_len = 0;
+    int container_hash_len = 0;
+    const char *container_hash;
+    const char *docker_id;
+    int name_found = FLB_FALSE;
+    /* Process status/containerStatus map for docker_id, container_hash */
+    for (i = 0;
+         (meta->docker_id_len == 0 || meta->container_hash_len == 0) &&
+         i < status.via.map.size; i++) {
+        k = status.via.map.ptr[i].key;
+        if (k.via.str.size == sizeof(CONTAINER_STATUSES)-1 &&
+            strncmp(k.via.str.ptr,
+                    CONTAINER_STATUSES,
+                    sizeof(CONTAINER_STATUSES)-1) == 0) {
+            int j;
+            v = status.via.map.ptr[i].val;
+            for (j = 0;
+                 (meta->docker_id_len == 0 ||
+                  meta->container_hash_len == 0) && j < v.via.array.size;
+                 j++) {
+                int l;
+                msgpack_object k1, k2;
+                msgpack_object_str v2;
+                k1 = v.via.array.ptr[j];
+                for (l = 0;
+                     (meta->docker_id_len == 0 ||
+                      meta->container_hash_len == 0) &&
+                     l < k1.via.map.size; l++) {
+                    k2 = k1.via.map.ptr[l].key;
+                    v2 = k1.via.map.ptr[l].val.via.str;
+                    if (k2.via.str.size == sizeof("name") -1 &&
+                        !strncmp(k2.via.str.ptr, "name", k2.via.str.size)) {
+                        if (v2.size == meta->container_name_len &&
+                            !strncmp(v2.ptr,
+                                     meta->container_name,
+                                     meta->container_name_len)) {
+                            name_found = FLB_TRUE;
+                        }
+                    }
+                    if (k2.via.str.size == sizeof("containerID") - 1 &&
+                        !strncmp(k2.via.str.ptr,
+                                 "containerID",
+                                 k2.via.str.size)) {
+                        docker_id = v2.ptr;
+                        docker_id_len = v2.size;
+                    }
+                    else if (k2.via.str.size == sizeof("imageID") - 1 &&
+                              !strncmp(v2.ptr,
+                                       "imageID",
+                                      v2.size)) {
+                        container_hash = v2.ptr;
+                        container_hash_len = v2.size;
+                    }
+                }
+                if (name_found) {
+                    if (container_hash_len && !meta->container_hash_len) {
+                        meta->container_hash_len = container_hash_len;
+                        meta->container_hash = flb_strndup(container_hash,
+                                                           container_hash_len);
+                        meta->skip++;
+                    }
+                    if (docker_id_len && !meta->docker_id_len) {
+                        meta->docker_id_len = docker_id_len;
+                        meta->docker_id = flb_strndup(docker_id, docker_id_len);
+                        meta->skip++;
+                    }
+                }
+            }
+        }
+    }
+}
+
 static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
-                      char *api_buf, size_t api_size,
+                      const char *api_buf, size_t api_size,
                       char **out_buf, size_t *out_size)
 {
     int i;
@@ -257,6 +392,7 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
     int map_size;
     int meta_found = FLB_FALSE;
     int spec_found = FLB_FALSE;
+    int status_found = FLB_FALSE;
     int have_uid = -1;
     int have_labels = -1;
     int have_annotations = -1;
@@ -271,6 +407,7 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
     msgpack_object v;
     msgpack_object meta_val;
     msgpack_object spec_val;
+    msgpack_object status_val;
     msgpack_object api_map;
     msgpack_object ann_map;
     struct flb_kube_props props = {0};
@@ -321,24 +458,24 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
      * }
      *
      * We are interested into the 'metadata' map value.
+     * We are also interested in the spec.nodeName.
+     * We are also interested in the status.containerStatuses.
      */
-    for (i = 0; i < api_map.via.map.size; i++) {
-        k = api_map.via.map.ptr[i].key;
-        if (k.via.str.size == 8 && strncmp(k.via.str.ptr, "metadata", 8) == 0) {
+    for (i = 0; !(meta_found && spec_found && status_found) &&
+                i < api_map.via.map.size; i++) {
+	k = api_map.via.map.ptr[i].key;
+        if (k.via.str.size == 8 && !strncmp(k.via.str.ptr, "metadata", 8)) {
             meta_val = api_map.via.map.ptr[i].val;
             meta_found = FLB_TRUE;
-            break;
         }
-    }
-
-    /* We are also interested in the nodeName from 'spec' map value. */
-    for (i = 0; i < api_map.via.map.size; i++) {
-       k = api_map.via.map.ptr[i].key;
-       if (k.via.str.size == 4 && strncmp(k.via.str.ptr, "spec", 4) == 0) {
+	else if (k.via.str.size == 4 && !strncmp(k.via.str.ptr, "spec", 4)) {
            spec_val = api_map.via.map.ptr[i].val;
            spec_found = FLB_TRUE;
-           break;
-       }
+        }
+	else if (k.via.str.size == 6 && !strncmp(k.via.str.ptr, "status", 6)) {
+           status_val = api_map.via.map.ptr[i].val;
+           status_found = FLB_TRUE;
+	}
     }
 
     if (meta_found == FLB_FALSE) {
@@ -352,7 +489,7 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
     for (i = 0; i < meta_val.via.map.size; i++) {
         k = meta_val.via.map.ptr[i].key;
 
-        char *ptr = (char *) k.via.str.ptr;
+        const char *ptr = k.via.str.ptr;
         size_t size = k.via.str.size;
 
         if (size == 3 && strncmp(ptr, "uid", 3) == 0) {
@@ -361,7 +498,9 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
         }
         else if (size == 6 && strncmp(ptr, "labels", 6) == 0) {
             have_labels = i;
-            map_size++;
+            if (ctx->labels == FLB_TRUE) {
+                map_size++;
+            }
         }
 
         else if (size == 11 && strncmp(ptr, "annotations", 11) == 0) {
@@ -389,6 +528,10 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
         }
     }
 
+    if ((!meta->container_hash || !meta->docker_id) && status_found) {
+        extract_container_hash(meta, status_val);
+    }
+
     /* Append Regex fields */
     msgpack_pack_map(&mp_pck, map_size);
     if (meta->podname != NULL) {
@@ -413,7 +556,7 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
         msgpack_pack_object(&mp_pck, v);
     }
 
-    if (have_labels >= 0) {
+    if (have_labels >= 0 && ctx->labels == FLB_TRUE) {
         k = meta_val.via.map.ptr[have_labels].key;
         v = meta_val.via.map.ptr[have_labels].val;
 
@@ -452,9 +595,9 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
 
                     /* Validate and set the property */
                     flb_kube_prop_set(ctx, meta,
-                                      (char *) k.via.str.ptr + 13,
+                                      k.via.str.ptr + 13,
                                       k.via.str.size - 13,
-                                      (char *) v.via.str.ptr,
+                                      v.via.str.ptr,
                                       v.via.str.size,
                                       &props);
                 }
@@ -480,14 +623,17 @@ static int merge_meta(struct flb_kube_meta *meta, struct flb_kube *ctx,
     return 0;
 }
 
-static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
-                               char *data, size_t data_size,
+static inline int extract_meta(struct flb_kube *ctx,
+                               const char *tag, int tag_len,
+                               const char *data, size_t data_size,
                                struct flb_kube_meta *meta)
 {
     int i;
     size_t off = 0;
     ssize_t n;
-    char *container = NULL;
+    int kube_tag_len;
+    const char *kube_tag_str;
+    const char *container = NULL;
     int container_found = FLB_FALSE;
     int container_length = 0;
     struct flb_regex_search result;
@@ -504,7 +650,7 @@ static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
     if (ctx->use_journal == FLB_TRUE) {
         off = 0;
         msgpack_unpacked_init(&mp_result);
-        while (msgpack_unpack_next(&mp_result, data, data_size, &off)) {
+        while (msgpack_unpack_next(&mp_result, data, data_size, &off) == MSGPACK_UNPACK_SUCCESS) {
             root = mp_result.data;
             if (root.type != MSGPACK_OBJECT_ARRAY) {
                 continue;
@@ -520,7 +666,7 @@ static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
 
                 if (strncmp(key.via.str.ptr, "CONTAINER_NAME", 14) == 0) {
                     val = map.via.map.ptr[i].val;
-                    container = (char *) val.via.str.ptr;
+                    container = val.via.str.ptr;
                     container_length = val.via.str.size;
                     container_found = FLB_TRUE;
                     break;
@@ -537,13 +683,26 @@ static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
             return -1;
         }
         n = flb_regex_do(ctx->regex,
-                         (unsigned char *) container, container_length,
+                         container, container_length,
                          &result);
         msgpack_unpacked_destroy(&mp_result);
     }
     else {
-        /* Lookup using Tag string */
-        n = flb_regex_do(ctx->regex, (unsigned char *) tag, tag_len, &result);
+        /*
+         * Lookup metadata using regular expression. In order to let the
+         * regex work we need to know before hand what's the Tag prefix
+         * set and make sure the adjustment can be done.
+         */
+        kube_tag_len = flb_sds_len(ctx->kube_tag_prefix);
+        if (kube_tag_len + 1 >= tag_len) {
+            flb_error("[filter_kube] incoming record tag (%s) is shorter "
+                      "than kube_tag_prefix value (%s)",
+                      tag, ctx->kube_tag_prefix);
+            return -1;
+        }
+        kube_tag_str = tag + kube_tag_len;
+        kube_tag_len = tag_len - kube_tag_len;
+        n = flb_regex_do(ctx->regex, kube_tag_str, kube_tag_len, &result);
     }
 
     if (n <= 0) {
@@ -556,8 +715,12 @@ static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
 
     /* Compose API server cache key */
     if (meta->podname && meta->namespace) {
-        meta->cache_key_len = meta->podname_len + meta->namespace_len + 1;
-        meta->cache_key = flb_malloc(meta->cache_key_len + 1);
+        /* calculate estimated buffer size */
+        n = meta->namespace_len + 1 + meta->podname_len + 1;
+        if (meta->container_name) {
+            n += meta->container_name_len + 1;
+        }
+        meta->cache_key = flb_malloc(n);
         if (!meta->cache_key) {
             flb_errno();
             return -1;
@@ -573,7 +736,16 @@ static inline int extract_meta(struct flb_kube *ctx, char *tag, int tag_len,
         /* Copy podname */
         memcpy(meta->cache_key + off, meta->podname, meta->podname_len);
         off += meta->podname_len;
+
+        if (meta->container_name) {
+            /* Separator */
+            meta->cache_key[off++] = ':';
+            memcpy(meta->cache_key + off, meta->container_name, meta->container_name_len);
+            off += meta->container_name_len;
+        }
+
         meta->cache_key[off] = '\0';
+        meta->cache_key_len = off;
     }
     else {
         meta->cache_key = NULL;
@@ -657,35 +829,6 @@ static int flb_kube_network_init(struct flb_kube *ctx, struct flb_config *config
     return 0;
 }
 
-static int flb_dummy_meta(char **out_buf, size_t *out_size)
-{
-    int len;
-    time_t t;
-    char stime[32];
-    struct tm result;
-    msgpack_sbuffer mp_sbuf;
-    msgpack_packer mp_pck;
-
-    t = time(NULL);
-    localtime_r(&t, &result);
-    asctime_r(&result, stime);
-    len = strlen(stime) - 1;
-
-    msgpack_sbuffer_init(&mp_sbuf);
-    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
-
-    msgpack_pack_map(&mp_pck, 1);
-    msgpack_pack_str(&mp_pck, 5 /* dummy */ );
-    msgpack_pack_str_body(&mp_pck, "dummy", 5);
-    msgpack_pack_str(&mp_pck, len);
-    msgpack_pack_str_body(&mp_pck, stime, len);
-
-    *out_buf = mp_sbuf.data;
-    *out_size = mp_sbuf.size;
-
-    return 0;
-}
-
 /* Initialize local context */
 int flb_kube_meta_init(struct flb_kube *ctx, struct flb_config *config)
 {
@@ -731,25 +874,50 @@ int flb_kube_meta_init(struct flb_kube *ctx, struct flb_config *config)
     return 0;
 }
 
+int flb_kube_dummy_meta_get(char **out_buf, size_t *out_size)
+{
+    int len;
+    time_t t;
+    char stime[32];
+    struct tm result;
+    msgpack_sbuffer mp_sbuf;
+    msgpack_packer mp_pck;
+
+    t = time(NULL);
+    localtime_r(&t, &result);
+    asctime_r(&result, stime);
+    len = strlen(stime) - 1;
+
+    msgpack_sbuffer_init(&mp_sbuf);
+    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
+
+    msgpack_pack_map(&mp_pck, 1);
+    msgpack_pack_str(&mp_pck, 5 /* dummy */ );
+    msgpack_pack_str_body(&mp_pck, "dummy", 5);
+    msgpack_pack_str(&mp_pck, len);
+    msgpack_pack_str_body(&mp_pck, stime, len);
+
+    *out_buf = mp_sbuf.data;
+    *out_size = mp_sbuf.size;
+
+    return 0;
+}
+
 int flb_kube_meta_get(struct flb_kube *ctx,
-                      char *tag, int tag_len,
-                      char *data, size_t data_size,
-                      char **out_buf, size_t *out_size,
+                      const char *tag, int tag_len,
+                      const char *data, size_t data_size,
+                      const char **out_buf, size_t *out_size,
                       struct flb_kube_meta *meta,
                       struct flb_kube_props *props)
 {
     int id;
     int ret;
-    char *hash_meta_buf;
+    const char *hash_meta_buf;
+    char *tmp_hash_meta_buf;
     size_t off = 0;
     size_t hash_meta_size;
     msgpack_unpacked result;
 
-    if (ctx->dummy_meta == FLB_TRUE) {
-        flb_dummy_meta(out_buf, out_size);
-        return 0;
-    }
-
     /* Get metadata from tag or record (cache key is the important one) */
     ret = extract_meta(ctx, tag, tag_len, data, data_size, meta);
     if (ret != 0) {
@@ -763,21 +931,21 @@ int flb_kube_meta_get(struct flb_kube *ctx,
     if (ret == -1) {
         /* Retrieve API server meta and merge with local meta */
         ret = get_and_merge_meta(ctx, meta,
-                                 &hash_meta_buf, &hash_meta_size);
+                                 &tmp_hash_meta_buf, &hash_meta_size);
         if (ret == -1) {
             return -1;
         }
 
         id = flb_hash_add(ctx->hash_table,
                           meta->cache_key, meta->cache_key_len,
-                          hash_meta_buf, hash_meta_size);
+                          tmp_hash_meta_buf, hash_meta_size);
         if (id >= 0) {
             /*
              * Release the original buffer created on extract_meta() as a new
              * copy have been generated into the hash table, then re-set
              * the outgoing buffer and size.
              */
-            flb_free(hash_meta_buf);
+            flb_free(tmp_hash_meta_buf);
             flb_hash_get_by_id(ctx->hash_table, id, meta->cache_key,
                                &hash_meta_buf, &hash_meta_size);
         }
