@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -24,12 +25,12 @@
 #include "kube_props.h"
 
 int flb_kube_prop_set(struct flb_kube *ctx, struct flb_kube_meta *meta,
-                      char *prop, int prop_len,
-                      char *val_buf, size_t val_len,
+                      const char *prop, int prop_len,
+                      const char *val_buf, size_t val_len,
                       struct flb_kube_props *props);
 int flb_kube_prop_pack(struct flb_kube_props *props,
                        void **out_buf, size_t *out_size);
-int flb_kube_prop_unpack(struct flb_kube_props *props, char *buf, size_t size);
+int flb_kube_prop_unpack(struct flb_kube_props *props, const char *buf, size_t size);
 void flb_kube_prop_destroy(struct flb_kube_props *props);
 
 #endif
