@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -24,12 +25,14 @@
 #include <fluent-bit/flb_sds.h>
 
 /* Property structure/array index */
-#define FLB_KUBE_PROPS_PARSER   0
-#define FLB_KUBE_PROPS_EXCLUDE  1
+#define FLB_KUBE_PROPS_STDOUT_PARSER   0
+#define FLB_KUBE_PROPS_STDERR_PARSER   1
+#define FLB_KUBE_PROPS_EXCLUDE  2
 
 struct flb_kube_props {
-    flb_sds_t parser;      /* suggested parser     */
-    int exclude;           /* bool: exclude logs ? */
+    flb_sds_t stdout_parser; /* suggested parser for stdout, and stderr if not spec */
+    flb_sds_t stderr_parser; /* suggested parser for stderr-only */
+    int exclude;             /* bool: exclude logs ? */
 };
 
 #endif
