@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -21,6 +22,7 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_pack.h>
 #include <fluent-bit/flb_parser.h>
+#include <fluent-bit/flb_unescape.h>
 
 #include "kube_conf.h"
 #include "kube_meta.h"
@@ -31,37 +33,76 @@
 #include <msgpack.h>
 
 /* Merge status used by merge_log_handler() */
-#define MERGE_UNESCAPED   0 /* merge unescaped string in temporal buffer */
+#define MERGE_NONE        0 /* merge unescaped string in temporal buffer */
 #define MERGE_PARSED      1 /* merge parsed string (log_buf)             */
-#define MERGE_BINARY      2 /* merge direct binary object (v)            */
+#define MERGE_MAP         2 /* merge direct binary object (v)            */
 
-static int unescape_string(char *buf, int buf_len, char **unesc_buf)
+#define T_LOG_STREAM "stream"
+#define T_LOG_STDERR "stderr"
+
+static int is_stream_stderr(const void *data, size_t bytes)
 {
-    int i = 0;
-    int j = 0;
-    char *p;
-    char n;
+    int i;
+    msgpack_unpacked result;
+    size_t off = 0;
+    msgpack_object root;
+    msgpack_object_map map;
+    msgpack_object k;
+    msgpack_object v;
+    msgpack_unpacked_init(&result);
+    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
+        root = result.data;
+        if (root.type != MSGPACK_OBJECT_ARRAY) {
+            continue;
+        }
 
-    p = *unesc_buf;
-    while (i < buf_len) {
-        if (buf[i] == '\\') {
-            if (i + 1 < buf_len) {
-                n = buf[i + 1];
-                if (n != 'a' && n != 'b' &&
-                    n != 't' && n != 'n' &&
-                    n != 'v' && n != 'f' &&
-                    n != 'r') {
-                    i++;
+        if (root.via.array.ptr[1].type != MSGPACK_OBJECT_MAP) {
+            continue;
+        }
+        map = root.via.array.ptr[1].via.map;
+        for (i = 0; i < map.size; i++) {
+            k = map.ptr[i].key;
+            v = map.ptr[i].val;
+
+            if (k.type == MSGPACK_OBJECT_STR) {
+                /* Validate 'log' field */
+                if (k.via.str.size == sizeof(T_LOG_STREAM)-1 &&
+                    strncmp(k.via.str.ptr, T_LOG_STREAM, sizeof(T_LOG_STREAM)-1) == 0) {
+                    if (!strncmp(v.via.str.ptr, T_LOG_STDERR, sizeof(T_LOG_STDERR)-1)) {
+                        msgpack_unpacked_destroy(&result);
+                        return 1;
+                    }
+                    break;
                 }
             }
-            else {
-                i++;
-            }
         }
-        p[j++] = buf[i++];
     }
-    p[j] = '\0';
-    return j;
+    msgpack_unpacked_destroy(&result);
+    return 0;
+}
+
+static int value_trim_size(msgpack_object o)
+{
+    int i;
+    int size = o.via.str.size;
+
+    for (i = size - 1; i > 0; i--) {
+        if (o.via.str.ptr[i] == '\n') {
+            size -= 1;
+            continue;
+        }
+
+        if (o.via.str.ptr[i - 1] == '\\' &&
+            (o.via.str.ptr[i] == 'n' || o.via.str.ptr[i] == 'r')) {
+            size -= 2;
+            i--;
+        }
+        else {
+            break;
+        }
+    }
+
+    return size;
 }
 
 static int merge_log_handler(msgpack_object o,
@@ -70,17 +111,14 @@ static int merge_log_handler(msgpack_object o,
                              struct flb_time *log_time,
                              struct flb_kube *ctx)
 {
-    int i;
     int ret;
-    int size;
     int new_size;
-    int unesc_len = 0;
+    int root_type;
     char *tmp;
 
     /* Reset vars */
     *out_buf = NULL;
     *out_size = 0;
-    ctx->unesc_buf_len = 0;
 
     /* Allocate more space if required */
     if (o.via.str.size >= ctx->unesc_buf_size) {
@@ -96,36 +134,16 @@ static int merge_log_handler(msgpack_object o,
         }
     }
 
-    /*
-     * Check where to cut the string if common ending bytes like \r or \n
-     * exists.
-     */
-    size = o.via.str.size;
-    for (i = size - 1; i > 0; i--) {
-        if (o.via.str.ptr[i] == '\n') {
-            size -= 1;
-            i--;
-            continue;
-        }
-
-        if (o.via.str.ptr[i - 1] == '\\' &&
-            (o.via.str.ptr[i] == 'n' || o.via.str.ptr[i] == 'r')) {
-            size -= 2;
-            i--;
-        }
-        else {
-            break;
-        }
-    }
-
-    /* Unescape application string */
-    unesc_len = unescape_string((char *) o.via.str.ptr,
-                                size, &ctx->unesc_buf);
-    ctx->unesc_buf_len = unesc_len;
+    /* Copy the string value and append the required NULL byte */
+    ctx->unesc_buf_len = (int) o.via.str.size;
+    memcpy(ctx->unesc_buf, o.via.str.ptr, o.via.str.size);
+    ctx->unesc_buf[ctx->unesc_buf_len] = '\0';
 
     ret = -1;
+
+    /* Parser set by Annotation */
     if (parser) {
-        ret = flb_parser_do(parser, ctx->unesc_buf, unesc_len,
+        ret = flb_parser_do(parser, ctx->unesc_buf, ctx->unesc_buf_len,
                             out_buf, out_size, log_time);
         if (ret >= 0) {
             if (flb_time_to_double(log_time) == 0) {
@@ -134,14 +152,31 @@ static int merge_log_handler(msgpack_object o,
             return MERGE_PARSED;
         }
     }
-    else {
-        ret = flb_pack_json(ctx->unesc_buf, unesc_len,
-                            (char **) out_buf, out_size);
+    else if (ctx->merge_parser) { /* Custom parser 'merge_parser' option */
+        ret = flb_parser_do(ctx->merge_parser,
+                            ctx->unesc_buf, ctx->unesc_buf_len,
+                            out_buf, out_size, log_time);
+        if (ret >= 0) {
+            if (flb_time_to_double(log_time) == 0) {
+                flb_time_get(log_time);
+            }
+            return MERGE_PARSED;
+        }
+    }
+    else { /* Default JSON parser */
+        ret = flb_pack_json(ctx->unesc_buf, ctx->unesc_buf_len,
+                            (char **) out_buf, out_size, &root_type);
+        if (ret == 0 && root_type != FLB_PACK_JSON_OBJECT) {
+            flb_debug("[filter_kube] could not merge JSON, root_type=%i",
+                      root_type);
+            flb_free(*out_buf);
+            return MERGE_NONE;
+        }
     }
 
     if (ret == -1) {
         flb_debug("[filter_kube] could not merge JSON log as requested");
-        return MERGE_UNESCAPED;
+        return MERGE_NONE;
     }
 
     return MERGE_PARSED;
@@ -183,13 +218,14 @@ static int cb_kube_init(struct flb_filter_instance *f_ins,
 
 static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
                             msgpack_object source_map,
-                            char *kube_buf, size_t kube_size,
+                            const char *kube_buf, size_t kube_size,
                             struct flb_kube_meta *meta,
+                            struct flb_time *time_lookup,
                             struct flb_parser *parser,
                             struct flb_kube *ctx)
 {
     int i;
-    int map_size;
+    int map_size = 0;
     int merge_status = -1;
     int new_map_size = 0;
     int log_index = -1;
@@ -201,7 +237,7 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
     msgpack_object k;
     msgpack_object v;
     msgpack_object root;
-    struct flb_time log_time = {0};
+    struct flb_time log_time;
 
     /* Original map size */
     map_size = source_map.via.map.size;
@@ -220,45 +256,56 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
         }
     }
 
-    if (flb_time_to_double(&log_time) == 0) {
-        flb_time_get(&log_time);
-    }
+    /* reset */
+    flb_time_zero(&log_time);
 
     /*
-     * If a log_index exists, the application log content inside the Docker JSON
-     * map is a escaped string. Proceed to reserve a temporal buffer and create
-     * an unescaped version.
+     * If a log_index exists, the application log content inside the
+     * Docker JSON map is a escaped string. Proceed to reserve a temporal
+     * buffer and create an unescaped version.
      */
     if (log_index != -1) {
         v = source_map.via.map.ptr[log_index].val;
-        if (v.type == MSGPACK_OBJECT_STR) {
-            merge_status = merge_log_handler(v, parser,
-                                             &log_buf, &log_size, &log_time,
-                                             ctx);
-        }
-        else if (v.type == MSGPACK_OBJECT_MAP) {
+        if (v.type == MSGPACK_OBJECT_MAP) {
             /* This is the easiest way, no extra processing required */
-            merge_status = MERGE_BINARY;
+            merge_status = MERGE_MAP;
+        }
+        else if (v.type == MSGPACK_OBJECT_STR) {
+            merge_status = merge_log_handler(v, parser,
+                                             &log_buf, &log_size,
+                                             &log_time, ctx);
+        }
+    }
+
+    /* Append record timestamp */
+    if (merge_status == MERGE_PARSED) {
+        if (flb_time_to_double(&log_time) == 0.0) {
+            flb_time_append_to_msgpack(time_lookup, pck, 0);
+        }
+        else {
+            flb_time_append_to_msgpack(&log_time, pck, 0);
         }
     }
     else {
-
+        flb_time_append_to_msgpack(time_lookup, pck, 0);
     }
 
     /* Determinate the size of the new map */
     new_map_size = map_size;
 
-    /* If a merged json exists, check the number of entries in that new map */
+    /* If a merged status exists, check the number of entries to merge */
     if (log_index != -1) {
         if (merge_status == MERGE_PARSED) {
             off = 0;
             msgpack_unpacked_init(&result);
             msgpack_unpack_next(&result, log_buf, log_size, &off);
             root = result.data;
-            log_buf_entries = root.via.map.size;
+            if (root.type == MSGPACK_OBJECT_MAP) {
+                log_buf_entries = root.via.map.size;
+            }
             msgpack_unpacked_destroy(&result);
         }
-        else if (merge_status == MERGE_BINARY) {
+        else if (merge_status == MERGE_MAP) {
             /* object 'v' represents the original binary log */
             log_buf_entries = v.via.map.size;
         }
@@ -269,14 +316,21 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
         new_map_size++;
     }
 
-    /* Start packaging the final map */
-    if (ctx->merge_json_key != NULL) {
-        /* Make room for one new key that will hold the original log entries */
-        new_map_size++;
+    if (log_buf_entries > 0) {
+        if (ctx->merge_log_key != NULL) {
+            new_map_size++;
+        }
+        else {
+            new_map_size += log_buf_entries;
+        }
     }
-    else {
-        new_map_size += log_buf_entries;
+
+    if ((merge_status == MERGE_PARSED || merge_status == MERGE_MAP) &&
+        ctx->keep_log == FLB_FALSE) {
+        new_map_size--;
     }
+
+    /* Pack Map */
     msgpack_pack_map(pck, new_map_size);
 
     /* Original map */
@@ -285,17 +339,28 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
         v = source_map.via.map.ptr[i].val;
 
         /*
-         * If the original 'log' field was unescaped and converted to
-         * msgpack properly, re-pack the new string version to avoid
-         * multiple escape sequences in outgoing plugins.
+         * If log_index is set, means a merge log is a requirement but
+         * will depend on merge_status. If the parsing failed we cannot
+         * merge so we keep the 'log' key/value.
          */
-        if (log_index == i &&
-            (merge_status == MERGE_UNESCAPED || merge_status == MERGE_PARSED)) {
-            msgpack_pack_object(pck, k);
-            msgpack_pack_str(pck, ctx->unesc_buf_len);
-            msgpack_pack_str_body(pck, ctx->unesc_buf, ctx->unesc_buf_len);
+        if (log_index == i) {
+            if (ctx->keep_log == FLB_TRUE) {
+                msgpack_pack_object(pck, k);
+                if (merge_status == MERGE_NONE || merge_status == MERGE_PARSED){
+                    msgpack_pack_str(pck, ctx->unesc_buf_len);
+                    msgpack_pack_str_body(pck, ctx->unesc_buf,
+                                          ctx->unesc_buf_len);
+                }
+                else {
+                    msgpack_pack_object(pck, v);
+                }
+            }
+            else if (merge_status == MERGE_NONE) {
+                msgpack_pack_object(pck, k);
+                msgpack_pack_object(pck, v);
+            }
         }
-        else { /* MERGE_BINARY ? */
+        else {
             msgpack_pack_object(pck, k);
             msgpack_pack_object(pck, v);
         }
@@ -304,10 +369,10 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
     /* Merge Log */
     if (log_index != -1) {
         if (merge_status == MERGE_PARSED) {
-            if (ctx->merge_json_key && log_buf_entries > 0) {
-                msgpack_pack_str(pck, ctx->merge_json_key_len);
-                msgpack_pack_str_body(pck, ctx->merge_json_key,
-                                      ctx->merge_json_key_len);
+            if (ctx->merge_log_key && log_buf_entries > 0) {
+                msgpack_pack_str(pck, ctx->merge_log_key_len);
+                msgpack_pack_str_body(pck, ctx->merge_log_key,
+                                      ctx->merge_log_key_len);
                 msgpack_pack_map(pck, log_buf_entries);
             }
 
@@ -315,21 +380,44 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
             msgpack_unpacked_init(&result);
             msgpack_unpack_next(&result, log_buf, log_size, &off);
             root = result.data;
+
             for (i = 0; i < log_buf_entries; i++) {
                 k = root.via.map.ptr[i].key;
-                v = root.via.map.ptr[i].val;
                 msgpack_pack_object(pck, k);
-                msgpack_pack_object(pck, v);
+
+                v = root.via.map.ptr[i].val;
+
+                /*
+                 * If this is the last string value, trim any remaining
+                 * break line or return carrier character.
+                 */
+                if (v.type == MSGPACK_OBJECT_STR &&
+                    ctx->merge_log_trim == FLB_TRUE) {
+                    int s = value_trim_size(v);
+                    msgpack_pack_str(pck, s);
+                    msgpack_pack_str_body(pck, v.via.str.ptr, s);
+                }
+                else {
+                    msgpack_pack_object(pck, v);
+                }
             }
             msgpack_unpacked_destroy(&result);
             flb_free(log_buf);
         }
-        else if (merge_status == MERGE_BINARY) {
-            msgpack_object bin_map;
-            bin_map = source_map.via.map.ptr[log_index].val;
-            for (i = 0; i < bin_map.via.map.size; i++) {
-                k = bin_map.via.map.ptr[i].key;
-                v = bin_map.via.map.ptr[i].val;
+        else if (merge_status == MERGE_MAP) {
+            msgpack_object map;
+
+            if (ctx->merge_log_key && log_buf_entries > 0) {
+                msgpack_pack_str(pck, ctx->merge_log_key_len);
+                msgpack_pack_str_body(pck, ctx->merge_log_key,
+                                      ctx->merge_log_key_len);
+                msgpack_pack_map(pck, log_buf_entries);
+            }
+
+            map = source_map.via.map.ptr[log_index].val;
+            for (i = 0; i < map.via.map.size; i++) {
+                k = map.via.map.ptr[i].key;
+                v = map.via.map.ptr[i].val;
                 msgpack_pack_object(pck, k);
                 msgpack_pack_object(pck, v);
             }
@@ -387,61 +475,75 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
     return 0;
 }
 
-static int cb_kube_filter(void *data, size_t bytes,
-                          char *tag, int tag_len,
+static int cb_kube_filter(const void *data, size_t bytes,
+                          const char *tag, int tag_len,
                           void **out_buf, size_t *out_bytes,
                           struct flb_filter_instance *f_ins,
                           void *filter_context,
                           struct flb_config *config)
 {
     int ret;
+    int is_stderr = 0;
     size_t pre = 0;
     size_t off = 0;
-    char *cache_buf = NULL;
+    char *dummy_cache_buf = NULL;
+    const char *cache_buf = NULL;
     size_t cache_size = 0;
     msgpack_unpacked result;
-    msgpack_object time;
     msgpack_object map;
     msgpack_object root;
     msgpack_sbuffer tmp_sbuf;
     msgpack_packer tmp_pck;
+    msgpack_object *obj;
     struct flb_parser *parser = NULL;
     struct flb_kube *ctx = filter_context;
     struct flb_kube_meta meta = {0};
     struct flb_kube_props props = {0};
+    struct flb_time time_lookup;
     (void) f_ins;
     (void) config;
 
-    if (ctx->use_journal == FLB_FALSE) {
-        /* Check if we have some cached metadata for the incoming events */
-        ret = flb_kube_meta_get(ctx,
-                                tag, tag_len,
-                                data, bytes,
-                                &cache_buf, &cache_size, &meta, &props);
+    if (ctx->use_journal == FLB_FALSE || ctx->dummy_meta == FLB_TRUE) {
+        if (ctx->dummy_meta == FLB_TRUE) {
+            ret = flb_kube_dummy_meta_get(&dummy_cache_buf, &cache_size);
+            cache_buf = dummy_cache_buf;
+        }
+        else {
+            /* Check if we have some cached metadata for the incoming events */
+            ret = flb_kube_meta_get(ctx,
+                                    tag, tag_len,
+                                    data, bytes,
+                                    &cache_buf, &cache_size, &meta, &props);
+        }
         if (ret == -1) {
             flb_kube_prop_destroy(&props);
             return FLB_FILTER_NOTOUCH;
         }
 
-        if (props.parser != NULL) {
-            parser = flb_parser_get(props.parser, config);
-        }
-
         if (props.exclude == FLB_TRUE) {
             *out_buf   = NULL;
             *out_bytes = 0;
+            flb_kube_meta_release(&meta);
             flb_kube_prop_destroy(&props);
             return FLB_FILTER_MODIFIED;
         }
     }
 
+    is_stderr = is_stream_stderr(data, bytes);
+    if (is_stderr && props.stderr_parser != NULL) {
+        parser = flb_parser_get(props.stderr_parser, config);
+    }
+    else if (props.stdout_parser != NULL) {
+        parser = flb_parser_get(props.stdout_parser, config);
+    }
+
     /* Create temporal msgpack buffer */
     msgpack_sbuffer_init(&tmp_sbuf);
     msgpack_packer_init(&tmp_pck, &tmp_sbuf, msgpack_sbuffer_write);
 
     /* Iterate each item array and append meta */
     msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off)) {
+    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
         root = result.data;
         if (root.type != MSGPACK_OBJECT_ARRAY) {
             continue;
@@ -454,7 +556,7 @@ static int cb_kube_filter(void *data, size_t bytes,
          * note: when the source is in_tail the situation is different since all
          * records passed to the filter have a unique source log file.
          */
-        if (ctx->use_journal == FLB_TRUE) {
+        if (ctx->use_journal == FLB_TRUE && ctx->dummy_meta == FLB_FALSE) {
             parser = NULL;
             cache_buf = NULL;
             memset(&props, '\0', sizeof(struct flb_kube_props));
@@ -470,10 +572,6 @@ static int cb_kube_filter(void *data, size_t bytes,
                 return FLB_FILTER_NOTOUCH;
             }
 
-            if (props.parser != NULL) {
-                parser = flb_parser_get(props.parser, config);
-            }
-
             if (props.exclude == FLB_TRUE) {
                 /* Skip this record */
                 continue;
@@ -482,23 +580,28 @@ static int cb_kube_filter(void *data, size_t bytes,
             pre = off;
         }
 
-        /* get time and map */
-        time = root.via.array.ptr[0];
+        /*
+         * Temporal time lookup in case a parser comes up with a new
+         * timestamp for the record.
+         */
+        flb_time_pop_from_msgpack(&time_lookup, &result, &obj);
+
+        /* get records map */
         map  = root.via.array.ptr[1];
 
-        /* Compose the new array */
+        /* Compose the new array (0=timestamp, 1=record) */
         msgpack_pack_array(&tmp_pck, 2);
-        msgpack_pack_object(&tmp_pck, time);
+
 
         ret = pack_map_content(&tmp_pck, &tmp_sbuf,
                                map,
                                cache_buf, cache_size,
-                               &meta, parser, ctx);
-        if (ret != 0) {
+                               &meta, &time_lookup, parser, ctx);
+        if (ret == -1) {
             msgpack_sbuffer_destroy(&tmp_sbuf);
             msgpack_unpacked_destroy(&result);
             if (ctx->dummy_meta == FLB_TRUE) {
-                flb_free(cache_buf);
+                flb_free(dummy_cache_buf);
             }
 
             flb_kube_meta_release(&meta);
@@ -522,7 +625,7 @@ static int cb_kube_filter(void *data, size_t bytes,
     *out_bytes = tmp_sbuf.size;
 
     if (ctx->dummy_meta == FLB_TRUE) {
-        flb_free(cache_buf);
+        flb_free(dummy_cache_buf);
     }
 
     flb_kube_prop_destroy(&props);
