@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -28,18 +29,24 @@ enum FILTER_NEST_OPERATION {
 struct filter_nest_ctx
 {
     enum FILTER_NEST_OPERATION operation;
+    char *key;
+    int key_len;
+    char *prefix;
+    int prefix_len;
     // nest
-    char *nesting_key;
-    int nesting_key_len;
-    char *wildcard;
-    int wildcard_len;
-    bool wildcard_is_dynamic;
+    struct mk_list wildcards;
+    int wildcards_cnt;
+    bool remove_prefix;
     // lift
-    char *nested_under;
-    int nested_under_len;
-    char *prefix_with;
-    int prefix_with_len;
-    bool use_prefix;
+    bool add_prefix;
+};
+
+struct filter_nest_wildcard
+{
+    char *key;
+    int key_len;
+    bool key_is_dynamic;
+    struct mk_list _head;
 };
 
 #endif
