@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -19,16 +20,15 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 
+#include <fluent-bit/flb_compat.h>
 #include <fluent-bit/flb_input.h>
 #include <fluent-bit/flb_config.h>
 #include <fluent-bit/flb_error.h>
 #include <fluent-bit/flb_utils.h>
-#include <fluent-bit/flb_stats.h>
 
 #include "tail.h"
 #include "tail_fs.h"
@@ -37,6 +37,7 @@
 #include "tail_scan.h"
 #include "tail_signal.h"
 #include "tail_config.h"
+#include "tail_dockermode.h"
 #include "tail_multiline.h"
 
 static inline int consume_byte(int fd)
@@ -45,7 +46,7 @@ static inline int consume_byte(int fd)
     uint64_t val;
 
     /* We need to consume the byte */
-    ret = read(fd, &val, sizeof(val));
+    ret = flb_pipe_r(fd, &val, sizeof(val));
     if (ret <= 0) {
         flb_errno();
         return -1;
@@ -137,6 +138,10 @@ static int in_tail_collect_static(struct flb_input_instance *i_ins,
             active++;
             continue;
         case FLB_TAIL_WAIT:
+            if (file->config->exit_on_eof) {
+                flb_info("[in_tail] file=%s ended, stop", file->name);
+                flb_engine_exit(config);
+            }
             /* Promote file to 'events' type handler */
             flb_debug("[in_tail] file=%s promote to TAIL_EVENT", file->name);
             ret = flb_tail_file_to_event(file);
@@ -248,6 +253,27 @@ static int in_tail_init(struct flb_input_instance *in,
     }
     ctx->coll_fd_pending = ret;
 
+
+    if (ctx->multiline == FLB_TRUE && ctx->parser) {
+        ctx->parser = NULL;
+        flb_warn("[in_tail] on multiline mode 'Parser' is not allowed "
+                 "(parser disabled)");
+    }
+
+    /* Register callback to process docker mode queued buffer */
+    if (ctx->docker_mode == FLB_TRUE) {
+        ret = flb_input_set_collector_time(in, flb_tail_dmode_pending_flush,
+                                           ctx->docker_mode_flush, 0,
+                                           config);
+        if (ret == -1) {
+            ctx->docker_mode = FLB_FALSE;
+            flb_tail_config_destroy(ctx);
+            return -1;
+        }
+        ctx->coll_fd_dmode_flush = ret;
+    }
+
+#ifdef FLB_HAVE_PARSER
     /* Register callback to process multiline queued buffer */
     if (ctx->multiline == FLB_TRUE) {
         ret = flb_input_set_collector_time(in, flb_tail_mult_pending_flush,
@@ -260,6 +286,7 @@ static int in_tail_init(struct flb_input_instance *in,
         }
         ctx->coll_fd_mult_flush = ret;
     }
+#endif
 
     return 0;
 }
@@ -301,6 +328,10 @@ static void in_tail_pause(void *data, struct flb_config *config)
     flb_input_collector_pause(ctx->coll_fd_static, ctx->i_ins);
     flb_input_collector_pause(ctx->coll_fd_pending, ctx->i_ins);
 
+    if (ctx->docker_mode == FLB_TRUE) {
+        flb_input_collector_pause(ctx->coll_fd_dmode_flush, ctx->i_ins);
+    }
+
     if (ctx->multiline == FLB_TRUE) {
         flb_input_collector_pause(ctx->coll_fd_mult_flush, ctx->i_ins);
     }
@@ -316,6 +347,10 @@ static void in_tail_resume(void *data, struct flb_config *config)
     flb_input_collector_resume(ctx->coll_fd_static, ctx->i_ins);
     flb_input_collector_resume(ctx->coll_fd_pending, ctx->i_ins);
 
+    if (ctx->docker_mode == FLB_TRUE) {
+        flb_input_collector_resume(ctx->coll_fd_dmode_flush, ctx->i_ins);
+    }
+
     if (ctx->multiline == FLB_TRUE) {
         flb_input_collector_resume(ctx->coll_fd_mult_flush, ctx->i_ins);
     }
