@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -24,11 +25,26 @@
 #include <fluent-bit/flb_input.h>
 #include <fluent-bit/flb_parser.h>
 #include <fluent-bit/flb_macros.h>
+#ifdef FLB_HAVE_REGEX
+#include <fluent-bit/flb_regex.h>
+#endif
+
+/* Metrics */
+#ifdef FLB_HAVE_METRICS
+#define FLB_TAIL_METRIC_F_OPENED  100  /* number of opened files  */
+#define FLB_TAIL_METRIC_F_CLOSED  101  /* number of closed files  */
+#define FLB_TAIL_METRIC_F_ROTATED 102  /* number of rotated files */
+#endif
 
 struct flb_tail_config {
     int fd_notify;             /* inotify fd               */
+#ifdef _WIN32
+    intptr_t ch_manager[2];    /* pipe: channel manager    */
+    intptr_t ch_pending[2];    /* pipe: pending events     */
+#else
     int ch_manager[2];         /* pipe: channel manager    */
     int ch_pending[2];         /* pipe: pending events     */
+#endif
 
     /* Buffer Config */
     size_t buf_chunk_size;     /* allocation chunks        */
@@ -39,6 +55,7 @@ struct flb_tail_config {
     int coll_fd_scan;
     int coll_fd_rotated;
     int coll_fd_pending;
+    int coll_fd_dmode_flush;
     int coll_fd_mult_flush;
 
     /* Backend collectors */
@@ -47,18 +64,22 @@ struct flb_tail_config {
 
     /* Configuration */
     int dynamic_tag;           /* dynamic tag ? e.g: abc.*     */
+#ifdef FLB_HAVE_REGEX
+    struct flb_regex *tag_regex;/* path to tag regex           */
+#endif
     int refresh_interval_sec;  /* seconds to re-scan           */
     long refresh_interval_nsec;/* nanoseconds to re-scan       */
     int rotate_wait;           /* sec to wait on rotated files */
     int ignore_older;          /* ignore fields older than X seconds        */
     time_t last_pending;       /* last time a 'pending signal' was emitted' */
-    char *path;                /* lookup path (glob)           */
-    char *exclude_path;        /* exclude path                 */
-    char *path_key;            /* key name of file path        */
+    const char *path;          /* lookup path (glob)           */
+    const char *exclude_path;  /* exclude path                 */
+    const char *path_key;      /* key name of file path        */
     int   path_key_len;        /* length of key name           */
     char *key;                 /* key for unstructured record  */
     int   key_len;             /* length of key ^              */
     int   skip_long_lines;     /* skip long lines              */
+    int   exit_on_eof;         /* exit fluent-bit on EOF, test */
 
     /* Database */
     struct flb_sqldb *db;
@@ -73,6 +94,10 @@ struct flb_tail_config {
     struct flb_parser *mult_parser_firstline;
     struct mk_list mult_parsers;
 
+    /* Docker mode */
+    int docker_mode;           /* Docker mode enabled ?  */
+    int docker_mode_flush;     /* Docker mode flush/wait */
+
     /* Lists head for files consumed statically (read) and by events (inotify) */
     struct mk_list files_static;
     struct mk_list files_event;
