@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -33,7 +34,7 @@ struct query_status {
 };
 
 /* Open or create database required by tail plugin */
-struct flb_sqldb *flb_tail_db_open(char *path,
+struct flb_sqldb *flb_tail_db_open(const char *path,
                                    struct flb_input_instance *in,
                                    struct flb_tail_config *ctx,
                                    struct flb_config *config)
@@ -94,6 +95,7 @@ int flb_tail_db_file_set(struct flb_tail_file *file,
     int ret;
     char query[PATH_MAX];
     struct query_status qs = {0};
+    uint64_t created;
 
     /* Check if the file exists */
     snprintf(query, sizeof(query) - 1,
@@ -106,9 +108,10 @@ int flb_tail_db_file_set(struct flb_tail_file *file,
 
     if (qs.rows == 0) {
         /* Register the file */
+        created = time(NULL);
         snprintf(query, sizeof(query) - 1,
                  SQL_INSERT_FILE,
-                 file->name, 0UL, file->inode, time(NULL));
+                 file->name, (uint64_t) 0, (uint64_t) file->inode, created);
         ret = flb_sqldb_query(ctx->db, query, NULL, NULL);
         if (ret == FLB_ERROR) {
             return -1;
@@ -133,7 +136,7 @@ int flb_tail_db_file_offset(struct flb_tail_file *file,
 
     snprintf(query, sizeof(query) - 1,
              SQL_UPDATE_OFFSET,
-             file->offset, file->db_id);
+             (uint64_t) file->offset, file->db_id);
 
     ret = flb_sqldb_query(ctx->db,
                           query, NULL, NULL);
@@ -145,7 +148,7 @@ int flb_tail_db_file_offset(struct flb_tail_file *file,
 }
 
 /* Mark a file as rotated */
-int flb_tail_db_file_rotate(char *new_name,
+int flb_tail_db_file_rotate(const char *new_name,
                             struct flb_tail_file *file,
                             struct flb_tail_config *ctx)
 {
