@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -20,7 +21,10 @@
 #ifndef FLB_TAIL_INTERNAL_H
 #define FLB_TAIL_INTERNAL_H
 
+#include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_input.h>
+#include <fluent-bit/flb_sds.h>
+#include <fluent-bit/flb_time.h>
 
 #include "tail.h"
 #include "tail_config.h"
@@ -34,8 +38,15 @@ struct flb_tail_file {
     off_t size;
     off_t offset;
     off_t last_line;
+#ifdef _MSC_VER
+    uint64_t inode;
+#else
     ino_t inode;
-    char *name;                 /* target file name */
+#endif
+    char *name;                 /* target file name given by scan routine */
+#if !defined(__linux)
+    char *real_name;            /* real file name in the file system */
+#endif
     size_t name_len;
     time_t rotated;
     off_t pending_bytes;
@@ -53,6 +64,11 @@ struct flb_tail_file {
     msgpack_packer mult_pck;    /* temporal msgpack packer               */
     struct flb_time mult_time;  /* multiline time parsed from first line */
 
+    /* docker mode */
+    time_t dmode_flush_timeout; /* time when docker mode started         */
+    flb_sds_t dmode_buf;        /* buffer for docker mode                */
+    flb_sds_t dmode_lastline;   /* last incomplete line                  */
+
     /* buffering */
     off_t parsed;
     off_t buf_len;
