@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -54,8 +55,4 @@ int flb_tail_mult_flush(msgpack_sbuffer *mp_sbuf,
 int flb_tail_mult_pending_flush(struct flb_input_instance *i_ins,
                                 struct flb_config *config, void *context);
 
-int flb_tail_file_pack_line(msgpack_sbuffer *mp_sbuf, msgpack_packer *mp_pck,
-                            struct flb_time *time, char *data, size_t data_size,
-                            struct flb_tail_file *file);
-
 #endif
