@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -42,10 +43,10 @@
 
 #define SQL_INSERT_FILE                                             \
     "INSERT INTO in_tail_files (name, offset, inode, created)"      \
-    "  VALUES ('%s', %lu, %lu, %lu);"
+    "  VALUES ('%s', %"PRIu64", %"PRIu64", %"PRIu64");"
 
 #define SQL_UPDATE_OFFSET                               \
-    "UPDATE in_tail_files set offset=%lu WHERE id=%"PRId64";"
+    "UPDATE in_tail_files set offset=%"PRIu64" WHERE id=%"PRId64";"
 
 #define SQL_ROTATE_FILE                         \
     "UPDATE in_tail_files set name='%s',rotated=1 WHERE id=%"PRId64";"
