@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -24,7 +25,8 @@ enum {
     FLB_OUT_FILE_FMT_JSON,
     FLB_OUT_FILE_FMT_CSV,
     FLB_OUT_FILE_FMT_LTSV,
-    FLB_OUT_FILE_FMT_OTHER,
+    FLB_OUT_FILE_FMT_PLAIN,
+    FLB_OUT_FILE_FMT_MSGPACK,
 };
 
 #endif
