@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -20,26 +21,20 @@
 #ifndef FLB_OUT_HTTP_H
 #define FLB_OUT_HTTP_H
 
-#define FLB_HTTP_OUT_MSGPACK        0
-#define FLB_HTTP_OUT_JSON           1
-#define FLB_HTTP_OUT_JSON_STREAM    2
-#define FLB_HTTP_OUT_JSON_LINES     3
-
-#define FLB_JSON_DATE_DOUBLE      0
-#define FLB_JSON_DATE_ISO8601     1
-#define FLB_JSON_DATE_ISO8601_FMT "%Y-%m-%dT%H:%M:%S"
+#define FLB_HTTP_OUT_MSGPACK        FLB_PACK_JSON_FORMAT_NONE
+#define FLB_HTTP_OUT_GELF           20
 
 #define FLB_HTTP_CONTENT_TYPE   "Content-Type"
 #define FLB_HTTP_MIME_MSGPACK   "application/msgpack"
 #define FLB_HTTP_MIME_JSON      "application/json"
 
-struct flb_out_http_config {
+struct flb_out_http {
     /* HTTP Auth */
     char *http_user;
     char *http_passwd;
 
     /* Proxy */
-    char *proxy;
+    const char *proxy;
     char *proxy_host;
     int proxy_port;
 
@@ -47,20 +42,34 @@ struct flb_out_http_config {
     int out_format;
 
     int json_date_format;
-    char *json_date_key;
-    size_t json_date_key_len;
+    flb_sds_t json_date_key;
 
     /* HTTP URI */
     char *uri;
     char *host;
     int port;
 
+    /* GELF fields */
+    struct flb_gelf_fields gelf_fields;
+
     /* Include tag in header */
     char *header_tag;
     size_t headertag_len;
 
     /* Upstream connection to the backend server */
     struct flb_upstream *u;
+
+    /* Arbitrary HTTP headers */
+    struct mk_list headers;
+    int headers_cnt;
+};
+
+struct out_http_header {
+    char *key;
+    int key_len;
+    char *val;
+    int val_len;
+    struct mk_list _head;
 };
 
 #endif
