@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -32,21 +33,37 @@
 /*
  * Returns FLB_TRUE if the specified value is true, otherwise FLB_FALSE
  */
-static int bool_value(char *v);
+static int bool_value(const char *v);
 
 /*
  * Returns FLB_TRUE when the specified key is in Tag_Keys list,
  * otherwise FLB_FALSE
  */
 static int is_tagged_key(struct flb_influxdb_config *ctx,
-                         char *key, int kl, int type);
+                         const char *key, int kl, int type);
+
+/*
+ * Increments the timestamp when it is duplicated
+ */
+static void influxdb_tsmod(struct flb_time *ts, struct flb_time *dupe,
+                           struct flb_time *last) {
+    if (flb_time_equal(ts, last) || flb_time_equal(ts, dupe)) {
+        ++dupe->tm.tv_nsec;
+        flb_time_copy(last, ts);
+        flb_time_copy(ts, dupe);
+    }
+    else {
+        flb_time_copy(last, ts);
+        flb_time_copy(dupe, ts);
+    }
+}
 
 /*
  * Convert the internal Fluent Bit data representation to the required one
  * by InfluxDB.
  */
-static char *influxdb_format(char *tag, int tag_len,
-                             void *data, size_t bytes, int *out_size,
+static char *influxdb_format(const char *tag, int tag_len,
+                             const void *data, size_t bytes, int *out_size,
                              struct flb_influxdb_config *ctx)
 {
     int i;
@@ -110,7 +127,7 @@ static char *influxdb_format(char *tag, int tag_len,
     off = 0;
     msgpack_unpacked_destroy(&result);
     msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off)) {
+    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
         if (result.data.type != MSGPACK_OBJECT_ARRAY) {
             continue;
         }
@@ -153,19 +170,19 @@ static char *influxdb_format(char *tag, int tag_len,
             int quote = FLB_FALSE;
 
             /* key */
-            char *key = NULL;
+            const char *key = NULL;
             int key_len;
 
             /* val */
-            char *val = NULL;
+            const char *val = NULL;
             int val_len;
 
             if (k->type == MSGPACK_OBJECT_STR) {
-                key = (char *) k->via.str.ptr;
+                key = k->via.str.ptr;
                 key_len = k->via.str.size;
             }
             else {
-                key = (char *) k->via.bin.ptr;
+                key = k->via.bin.ptr;
                 key_len = k->via.bin.size;
             }
 
@@ -199,13 +216,13 @@ static char *influxdb_format(char *tag, int tag_len,
             else if (v->type == MSGPACK_OBJECT_STR) {
                 /* String value */
                 quote   = FLB_TRUE;
-                val     = (char *) v->via.str.ptr;
+                val     = v->via.str.ptr;
                 val_len = v->via.str.size;
             }
             else if (v->type == MSGPACK_OBJECT_BIN) {
                 /* Bin value */
                 quote   = FLB_TRUE;
-                val     = (char *) v->via.bin.ptr;
+                val     = v->via.bin.ptr;
                 val_len = v->via.bin.size;
             }
 
@@ -231,7 +248,7 @@ static char *influxdb_format(char *tag, int tag_len,
                 ret = influxdb_bulk_append_kv(bulk_head,
                                               key, key_len,
                                               val, val_len,
-                                              quote);
+                                              false);
             }
             else {
                 /* Append key/value data into the bulk_body */
@@ -254,6 +271,8 @@ static char *influxdb_format(char *tag, int tag_len,
 
         /* Check have data fields */
         if (bulk_body->len > 0) {
+            /* Modify timestamp in avoidance of duplication */
+            influxdb_tsmod(&tm, &ctx->ts_dupe, &ctx->ts_last);
             /* Append the timestamp */
             ret = influxdb_bulk_append_timestamp(bulk_body, &tm);
             if (ret == -1) {
@@ -311,7 +330,7 @@ int cb_influxdb_init(struct flb_output_instance *ins, struct flb_config *config,
                      void *data)
 {
     int io_flags = 0;
-    char *tmp;
+    const char *tmp;
     struct flb_upstream *upstream;
     struct flb_influxdb_config *ctx;
 
@@ -353,6 +372,9 @@ int cb_influxdb_init(struct flb_output_instance *ins, struct flb_config *config,
     if (!tmp) {
         ctx->seq_name = flb_strdup("_seq");
     }
+    else if (strcmp(tmp, "off") == 0) {
+        ctx->seq_name = flb_strdup("");
+    }
     else {
         ctx->seq_name = flb_strdup(tmp);
     }
@@ -409,14 +431,17 @@ int cb_influxdb_init(struct flb_output_instance *ins, struct flb_config *config,
     ctx->u   = upstream;
     ctx->seq = 0;
 
+    flb_time_zero(&ctx->ts_dupe);
+    flb_time_zero(&ctx->ts_last);
+
     flb_debug("[out_influxdb] host=%s port=%i", ins->host.name, ins->host.port);
     flb_output_set_context(ins, ctx);
 
     return 0;
 }
 
-void cb_influxdb_flush(void *data, size_t bytes,
-                       char *tag, int tag_len,
+void cb_influxdb_flush(const void *data, size_t bytes,
+                       const char *tag, int tag_len,
                        struct flb_input_instance *i_ins,
                        void *out_context,
                        struct flb_config *config)
@@ -456,11 +481,11 @@ void cb_influxdb_flush(void *data, size_t bytes,
         if (c->resp.status != 200 && c->resp.status != 204) {
             if (c->resp.payload_size > 0) {
                 flb_error("[out_influxdb] http_status=%i\n%s",
-                          ret, c->resp.status, c->resp.payload);
+                          c->resp.status, c->resp.payload);
             }
             else {
                 flb_debug("[out_influxdb] http_status=%i",
-                          ret, c->resp.status);
+                          c->resp.status);
             }
         }
         flb_debug("[out_influxdb] http_do=%i OK", ret);
@@ -501,7 +526,7 @@ int cb_influxdb_exit(void *data, struct flb_config *config)
     return 0;
 }
 
-int bool_value(char *v)
+int bool_value(const char *v)
 {
     if (strcasecmp(v, "true") == 0) {
         return FLB_TRUE;
@@ -516,7 +541,7 @@ int bool_value(char *v)
     return FLB_FALSE;
 }
 
-int is_tagged_key(struct flb_influxdb_config *ctx, char *key, int kl, int type)
+int is_tagged_key(struct flb_influxdb_config *ctx, const char *key, int kl, int type)
 {
     if (type == MSGPACK_OBJECT_STR) {
         if (ctx->auto_tags) {
