@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -30,7 +31,7 @@ struct flb_kafka *flb_kafka_conf_create(struct flb_output_instance *ins,
                                         struct flb_config *config)
 {
     int ret;
-    char *tmp;
+    const char *tmp;
     char errstr[512];
     struct mk_list *head;
     struct mk_list *topics;
@@ -121,6 +122,9 @@ struct flb_kafka *flb_kafka_conf_create(struct flb_output_instance *ins,
         else if (strcasecmp(tmp, "msgpack") == 0) {
             ctx->format = FLB_KAFKA_FMT_MSGP;
         }
+        else if (strcasecmp(tmp, "gelf") == 0) {
+            ctx->format = FLB_KAFKA_FMT_GELF;
+        }
     }
     else {
         ctx->format = FLB_KAFKA_FMT_JSON;
@@ -148,6 +152,45 @@ struct flb_kafka *flb_kafka_conf_create(struct flb_output_instance *ins,
         ctx->timestamp_key_len = strlen(FLB_KAFKA_TS_KEY);
     }
 
+    /* Config: Timestamp_Format */
+    ctx->timestamp_format = FLB_JSON_DATE_DOUBLE;
+    tmp = flb_output_get_property("timestamp_format", ins);
+    if (tmp) {
+        if (strcasecmp(tmp, "iso8601") == 0) {
+            ctx->timestamp_format = FLB_JSON_DATE_ISO8601;
+        }
+    }
+
+    /* Config Gelf_Timestamp_Key */
+    tmp = flb_output_get_property("gelf_timestamp_key", ins);
+    if (tmp) {
+        ctx->gelf_fields.timestamp_key = flb_sds_create(tmp);
+    }
+
+    /* Config Gelf_Host_Key */
+    tmp = flb_output_get_property("gelf_host_key", ins);
+    if (tmp) {
+        ctx->gelf_fields.host_key = flb_sds_create(tmp);
+    }
+
+    /* Config Gelf_Short_Message_Key */
+    tmp = flb_output_get_property("gelf_short_message_key", ins);
+    if (tmp) {
+        ctx->gelf_fields.short_message_key = flb_sds_create(tmp);
+    }
+
+    /* Config Gelf_Full_Message_Key */
+    tmp = flb_output_get_property("gelf_full_message_key", ins);
+    if (tmp) {
+        ctx->gelf_fields.full_message_key = flb_sds_create(tmp);
+    }
+
+    /* Config Gelf_Level_Key */
+    tmp = flb_output_get_property("gelf_level_key", ins);
+    if (tmp) {
+        ctx->gelf_fields.level_key = flb_sds_create(tmp);
+    }
+
     /* Kafka Producer */
     ctx->producer = rd_kafka_new(RD_KAFKA_PRODUCER, ctx->conf,
                                  errstr, sizeof(errstr));
@@ -211,6 +254,12 @@ int flb_kafka_conf_destroy(struct flb_kafka *ctx)
         flb_free(ctx->message_key);
     }
 
+    flb_sds_destroy(ctx->gelf_fields.timestamp_key);
+    flb_sds_destroy(ctx->gelf_fields.host_key);
+    flb_sds_destroy(ctx->gelf_fields.short_message_key);
+    flb_sds_destroy(ctx->gelf_fields.full_message_key);
+    flb_sds_destroy(ctx->gelf_fields.level_key);
+
     flb_free(ctx);
     return 0;
 }
