@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -38,6 +39,9 @@ struct flb_splunk {
     /* Token Auth */
     flb_sds_t auth_header;
 
+    /* Send fields directly or pack data into "event" object */
+    int splunk_send_raw;
+
     /* Upstream connection to the backend server */
     struct flb_upstream *u;
 };
