@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -17,6 +18,9 @@
  *  limitations under the License.
  */
 
+#include <fluent-bit/flb_info.h>
+#include <fluent-bit/flb_utils.h>
+
 #include "splunk.h"
 #include "splunk_conf.h"
 
@@ -24,7 +28,7 @@ struct flb_splunk *flb_splunk_conf_create(struct flb_output_instance *ins,
                                           struct flb_config *config)
 {
     int io_flags = 0;
-    char *tmp;
+    const char *tmp;
     flb_sds_t t;
     struct flb_upstream *upstream;
     struct flb_splunk *ctx;
@@ -110,6 +114,15 @@ struct flb_splunk *flb_splunk_conf_create(struct flb_output_instance *ins,
         }
     }
 
+    /* Event format, send all fields or pack into event map */
+    tmp = flb_output_get_property("splunk_send_raw", ins);
+    if (tmp) {
+        ctx->splunk_send_raw = flb_utils_bool(tmp);
+    }
+    else {
+        ctx->splunk_send_raw = FLB_FALSE;
+    }
+
     return ctx;
 }
 
