@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -23,12 +24,15 @@
 #include <fluent-bit.h>
 #include <fluent-bit/flb_io.h>
 
+#define FLB_TD_REGION_US    0
+#define FLB_TD_REGION_JP    1
+
 struct flb_out_td_config {
     int fd;           /* Socket to destination/backend */
-
-    char *api;
-    char *db_name;
-    char *db_table;
+    int region;       /* TD Region end-point */
+    const char *api;
+    const char *db_name;
+    const char *db_table;
 
     struct flb_upstream *u;
 };
