@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -21,6 +22,7 @@
 
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_input.h>
+#include <fluent-bit/flb_input_chunk.h>
 #include <fluent-bit/flb_output.h>
 #include <fluent-bit/flb_router.h>
 #include <fluent-bit/flb_config.h>
@@ -28,15 +30,11 @@
 #include <fluent-bit/flb_engine.h>
 #include <fluent-bit/flb_task.h>
 
-void flb_task_add_thread(struct flb_thread *thread,
-                                struct flb_task *task);
-
-#if defined (FLB_HAVE_FLUSH_LIBCO)
-
 /* It creates a new output thread using a 'Retry' context */
 int flb_engine_dispatch_retry(struct flb_task_retry *retry,
                               struct flb_config *config)
 {
+    size_t buf_size;
     struct flb_thread *th;
     struct flb_task *task;
     struct flb_input_instance *i_ins;
@@ -44,13 +42,19 @@ int flb_engine_dispatch_retry(struct flb_task_retry *retry,
     task = retry->parent;
     i_ins = task->i_ins;
 
+    /* Set file up/down based on restrictions */
+    flb_input_chunk_set_up(task->ic);
+
+    /* There is a match, get the buffer */
+    task->buf = flb_input_chunk_flush(task->ic, &buf_size);
+    task->size = buf_size;
+
     th = flb_output_thread(task,
                            i_ins,
                            retry->o_ins,
                            config,
                            task->buf, task->size,
-                           task->tag,
-                           strlen(task->tag));
+                           task->tag, task->tag_len);
     if (!th) {
         return -1;
     }
@@ -95,7 +99,7 @@ static int tasks_start(struct flb_input_instance *in,
                                    config,
                                    task->buf, task->size,
                                    task->tag,
-                                   strlen(task->tag));
+                                   task->tag_len);
             flb_task_add_thread(th, task);
             flb_thread_resume(th);
         }
@@ -107,16 +111,23 @@ static int tasks_start(struct flb_input_instance *in,
 /*
  * The engine dispatch is responsible for:
  *
- * - Get records from input plugins (fixed tags and dynamic tags)
+ * - Get chunks generated by input plugins.
  * - For each set of records under the same tag, create a Task. A Task set
  *   a reference to the records and routes through output instances.
  */
 int flb_engine_dispatch(uint64_t id, struct flb_input_instance *in,
                         struct flb_config *config)
 {
-    char *buf;
-    size_t size;
+    int ret;
+    int t_err;
+    const char *buf_data;
+    size_t buf_size = 0;
+    const char *tag_buf;
+    int tag_len;
+    struct mk_list *tmp;
+    struct mk_list *head;
     struct flb_input_plugin *p;
+    struct flb_input_chunk *ic;
     struct flb_task *task = NULL;
 
     p = in->p;
@@ -124,90 +135,55 @@ int flb_engine_dispatch(uint64_t id, struct flb_input_instance *in,
         return 0;
     }
 
-    if (in->flags & FLB_INPUT_DYN_TAG) {
-        /* Iterate dynamic tag buffers */
-        struct mk_list *d_head, *tmp;
-        struct flb_input_dyntag *dt;
-
-        mk_list_foreach_safe(d_head, tmp, &in->dyntags) {
-            dt = mk_list_entry(d_head, struct flb_input_dyntag, _head);
-            if (dt->busy == FLB_TRUE) {
-                continue;
-            }
-
-            /* There is a match, get the buffer */
-            buf = flb_input_dyntag_flush(dt, &size);
-            if (size == 0) {
-                /*
-                 * Do not release the buffer since if allocated, it will be
-                 * released upon input_dyntag context destroy.
-                 */
-                continue;
-            }
-            if (!buf) {
-                continue;
-            }
-
-            flb_trace("[dyntag %s] %p tag=%s", dt->in->name, dt, dt->tag);
-            task = flb_task_create(id, buf, size, dt->in, dt, dt->tag, config);
-            if (!task) {
-                /* Do not release the buffer, will happen on dyntag destroy */
-                continue;
-            }
-        }
-    }
-    else {
-        /* Get data from instance buffers */
-        buf = flb_input_flush(in, &size);
-        if (!buf || size == 0) {
-            if (buf) {
-                flb_free(buf);
-            }
-            return 0;
+    /* Look for chunks ready to go */
+    mk_list_foreach_safe(head, tmp, &in->chunks) {
+        ic = mk_list_entry(head, struct flb_input_chunk, _head);
+        if (ic->busy == FLB_TRUE) {
+            continue;
         }
 
-        /*
-         * Create an engine task, the task will hold the buffer reference
-         * and the co-routines associated to the output instance plugins
-         * that needs to handle the data.
-         */
-        task = flb_task_create(id, buf, size, in, NULL, in->tag, config);
+        /* There is a match, get the buffer */
+        buf_data = flb_input_chunk_flush(ic, &buf_size);
+        if (buf_size == 0) {
+            /*
+             * Do not release the buffer since if allocated, it will be
+             * released when the task is destroyed.
+             */
+            flb_input_chunk_release_lock(ic);
+            continue;
+        }
+        if (!buf_data) {
+            flb_input_chunk_release_lock(ic);
+            continue;
+        }
+
+        /* Get the the tag reference (chunk metadata) */
+        ret = flb_input_chunk_get_tag(ic, &tag_buf, &tag_len);
+        if (ret == -1) {
+            flb_input_chunk_release_lock(ic);
+            continue;
+        }
+
+        /* Create a task */
+        task = flb_task_create(id, buf_data, buf_size,
+                               ic->in, ic,
+                               tag_buf, tag_len,
+                               config, &t_err);
         if (!task) {
-            flb_free(buf);
-            return -1;
+            /*
+             * If task creation failed, check the error status flag. An error
+             * is associated with memory allocation or exhaustion of tasks_id,
+             * on that case the input chunk must be preserved and retried
+             * later. So we just release it busy lock.
+             */
+            if (t_err == FLB_TRUE) {
+                flb_input_chunk_release_lock(ic);
+            }
+            continue;
         }
-        flb_trace("[engine dispatch] task #%i created %p", task->id, task);
     }
 
     /* Start the new enqueued Tasks */
     tasks_start(in, config);
     return 0;
 }
-
-/*
- * Given an input instance, buffer and a bitmask of routes, create the task
- * and routes associated for processing. This mechanism does direct routing
- * without the use of a Tag.
- */
-int flb_engine_dispatch_direct(uint64_t id,
-                               struct flb_input_instance *in,
-                               char *buf, size_t size,
-                               char *tag, uint64_t routes,
-                               char *hash_str,
-                               struct flb_config *config)
-{
-    struct flb_task *task;
-
-    task = flb_task_create_direct(id, buf, size, in, tag, hash_str,
-                                  routes, config);
-    if (!task) {
-        return -1;
-    }
-    flb_trace("[engine dispatch direct] task created %p", task);
-
-    /* Start the new enqueued Tasks */
-    tasks_start(in, config);
-    return 0;
-}
-
-#endif /* !FLB_HAVE_FLUSH_LIBCO */
