@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -52,10 +53,10 @@ static unsigned int gen_hash(const void *key, int len)
     uint32_t h = seed ^ len;
 
     /* Mix 4 bytes at a time into the hash */
-    const unsigned char *data = (const unsigned char *)key;
+    const unsigned char *data = (const unsigned char *) key;
 
-    while(len >= 4) {
-        uint32_t k = *(uint32_t*) data;
+    while (len >= 4) {
+        uint32_t k = *(uint32_t *) data;
 
         k *= m;
         k ^= k >> r;
@@ -69,10 +70,14 @@ static unsigned int gen_hash(const void *key, int len)
     }
 
     /* Handle the last few bytes of the input array  */
-    switch(len) {
-    case 3: h ^= data[2] << 16;
-    case 2: h ^= data[1] << 8;
-    case 1: h ^= data[0]; h *= m;
+    switch (len) {
+    case 3:
+        h ^= data[2] << 16;
+    case 2:
+        h ^= data[1] << 8;
+    case 1:
+        h ^= data[0];
+        h *= m;
     };
 
     /* Do a few final mixes of the hash to ensure the last few
@@ -84,7 +89,8 @@ static unsigned int gen_hash(const void *key, int len)
     return (unsigned int) h;
 }
 
-static inline void flb_hash_entry_free(struct flb_hash *ht, struct flb_hash_entry *entry)
+static inline void flb_hash_entry_free(struct flb_hash *ht,
+                                       struct flb_hash_entry *entry)
 {
     mk_list_del(&entry->_head);
     mk_list_del(&entry->_head_parent);
@@ -173,8 +179,9 @@ static void flb_hash_evict_random(struct flb_hash *ht)
     }
 }
 
-int flb_hash_add(struct flb_hash *ht, char *key, int key_len,
-                 char *val, size_t val_size)
+int flb_hash_add(struct flb_hash *ht,
+                 const char *key, int key_len,
+                 const char *val, size_t val_size)
 {
     int id;
     unsigned int hash;
@@ -219,7 +226,7 @@ int flb_hash_add(struct flb_hash *ht, char *key, int key_len,
     entry->hits = 0;
 
     /* Store the key and value as a new memory region */
-    entry->key = flb_strdup(key);
+    entry->key = flb_strndup(key, key_len);
     entry->key_len = key_len;
     entry->val = flb_malloc(val_size + 1);
     if (!entry->val) {
@@ -264,8 +271,9 @@ int flb_hash_add(struct flb_hash *ht, char *key, int key_len,
     return id;
 }
 
-int flb_hash_get(struct flb_hash *ht, char *key, int key_len,
-                 char **out_buf, size_t *out_size)
+int flb_hash_get(struct flb_hash *ht,
+                 const char *key, int key_len,
+                 const char **out_buf, size_t * out_size)
 {
     int id;
     unsigned int hash;
@@ -287,10 +295,10 @@ int flb_hash_get(struct flb_hash *ht, char *key, int key_len,
 
     if (table->count == 1) {
         entry = mk_list_entry_first(&table->chains,
-                                    struct flb_hash_entry,
-                                    _head);
+                                    struct flb_hash_entry, _head);
 
-        if (strcmp(entry->key, key) != 0) {
+        if (entry->key_len != key_len
+            || strncmp(entry->key, key, key_len) != 0) {
             entry = NULL;
         }
     }
@@ -303,7 +311,7 @@ int flb_hash_get(struct flb_hash *ht, char *key, int key_len,
                 continue;
             }
 
-            if (strcmp(entry->key, key) == 0) {
+            if (strncmp(entry->key, key, key_len) == 0) {
                 break;
             }
 
@@ -330,8 +338,9 @@ int flb_hash_get(struct flb_hash *ht, char *key, int key_len,
  * Get an entry based in the table id. Note that a table id might have multiple
  * entries so the 'key' parameter is required to get an exact match.
  */
-int flb_hash_get_by_id(struct flb_hash *ht, int id, char *key, char **out_buf,
-                       size_t *out_size)
+int flb_hash_get_by_id(struct flb_hash *ht, int id,
+                       const char *key,
+                       const char **out_buf, size_t * out_size)
 {
     struct mk_list *head;
     struct flb_hash_entry *entry = NULL;
@@ -366,7 +375,7 @@ int flb_hash_get_by_id(struct flb_hash *ht, int id, char *key, char **out_buf,
     return 0;
 }
 
-int flb_hash_del(struct flb_hash *ht, char *key)
+int flb_hash_del(struct flb_hash *ht, const char *key)
 {
     int id;
     int len;
@@ -392,6 +401,9 @@ int flb_hash_del(struct flb_hash *ht, char *key)
         entry = mk_list_entry_first(&table->chains,
                                     struct flb_hash_entry,
                                     _head);
+        if (strcmp(entry->key, key) != 0) {
+            entry = NULL;
+        }
     }
     else {
         mk_list_foreach(head, &table->chains) {
