@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -31,10 +32,11 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_engine.h>
 #include <fluent-bit/flb_metrics.h>
+#include <fluent-bit/flb_storage.h>
 
 #define protcmp(a, b)  strncasecmp(a, b, strlen(a))
 
-static int check_protocol(char *prot, char *output)
+static int check_protocol(const char *prot, const char *output)
 {
     int len;
 
@@ -59,7 +61,7 @@ static inline int instance_id(struct flb_input_plugin *p,
 
     mk_list_foreach(head, &config->inputs) {
         entry = mk_list_entry(head, struct flb_input_instance, _head);
-        if (entry->p == p) {
+        if (entry->id == c) {
             c++;
         }
     }
@@ -85,7 +87,8 @@ static int collector_id(struct flb_input_instance *in)
 
 /* Create an input plugin instance */
 struct flb_input_instance *flb_input_new(struct flb_config *config,
-                                         char *input, void *data)
+                                         const char *input, void *data,
+                                         int public_only)
 {
     int id;
     int ret;
@@ -104,10 +107,18 @@ struct flb_input_instance *flb_input_new(struct flb_config *config,
             continue;
         }
 
+        /*
+         * Check if the plugin is private and validate the 'public_only'
+         * requirement.
+         */
+        if (public_only == FLB_TRUE && plugin->flags & FLB_INPUT_PRIVATE) {
+            return NULL;
+        }
+
         /* Create plugin instance */
         instance = flb_malloc(sizeof(struct flb_input_instance));
         if (!instance) {
-            perror("malloc");
+            flb_errno();
             return NULL;
         }
         instance->config = config;
@@ -119,13 +130,17 @@ struct flb_input_instance *flb_input_new(struct flb_config *config,
         snprintf(instance->name, sizeof(instance->name) - 1,
                  "%s.%i", plugin->name, id);
 
+        instance->alias    = NULL;
         instance->id       = id;
         instance->flags    = plugin->flags;
         instance->p        = plugin;
         instance->tag      = NULL;
+        instance->tag_len  = 0;
+        instance->routable = FLB_TRUE;
         instance->context  = NULL;
         instance->data     = data;
         instance->threaded = FLB_FALSE;
+        instance->storage  = NULL;
 
         /* net */
         instance->host.name    = NULL;
@@ -133,21 +148,10 @@ struct flb_input_instance *flb_input_new(struct flb_config *config,
         instance->host.uri     = NULL;
         instance->host.ipv6    = FLB_FALSE;
 
-        /* Initialize msgpack counter and buffers */
-        instance->mp_records = 0;
-        msgpack_sbuffer_init(&instance->mp_sbuf);
-        msgpack_packer_init(&instance->mp_pck, &instance->mp_sbuf,
-                            msgpack_sbuffer_write);
-        instance->mp_zone = msgpack_zone_new(MSGPACK_ZONE_CHUNK_SIZE);
-        if (!instance->mp_zone) {
-            flb_free(instance);
-            return NULL;
-        }
-
         /* Initialize list heads */
         mk_list_init(&instance->routes);
         mk_list_init(&instance->tasks);
-        mk_list_init(&instance->dyntags);
+        mk_list_init(&instance->chunks);
         mk_list_init(&instance->properties);
         mk_list_init(&instance->collectors);
         mk_list_init(&instance->threads);
@@ -167,24 +171,17 @@ struct flb_input_instance *flb_input_new(struct flb_config *config,
         }
 
         instance->mp_total_buf_size = 0;
-        instance->mp_buf_limit = 0;
-        instance->mp_buf_status = FLB_INPUT_RUNNING;
+        instance->mem_buf_status = FLB_INPUT_RUNNING;
+        instance->mem_buf_limit = 0;
+        instance->mem_chunks_size = 0;
 
-        /* Metrics */
-#ifdef FLB_HAVE_METRICS
-        instance->metrics = flb_metrics_create(instance->name);
-        if (instance->metrics) {
-            flb_metrics_add(FLB_METRIC_N_RECORDS, "records", instance->metrics);
-            flb_metrics_add(FLB_METRIC_N_BYTES, "bytes", instance->metrics);
-        }
-#endif
         mk_list_add(&instance->_head, &config->inputs);
     }
 
     return instance;
 }
 
-static inline int prop_key_check(char *key, char *kv, int k_len)
+static inline int prop_key_check(const char *key, const char *kv, int k_len)
 {
     int len;
 
@@ -198,7 +195,8 @@ static inline int prop_key_check(char *key, char *kv, int k_len)
 }
 
 /* Override a configuration property for the given input_instance plugin */
-int flb_input_set_property(struct flb_input_instance *in, char *k, char *v)
+int flb_input_set_property(struct flb_input_instance *in,
+                           const char *k, const char *v)
 {
     int len;
     ssize_t limit;
@@ -219,13 +217,20 @@ int flb_input_set_property(struct flb_input_instance *in, char *k, char *v)
         in->tag     = tmp;
         in->tag_len = strlen(tmp);
     }
+    else if (prop_key_check("routable", k, len) == 0 && tmp) {
+        in->routable = flb_utils_bool(tmp);
+        flb_free(tmp);
+    }
+    else if (prop_key_check("alias", k, len) == 0 && tmp) {
+        in->alias = tmp;
+    }
     else if (prop_key_check("mem_buf_limit", k, len) == 0 && tmp) {
         limit = flb_utils_size_to_bytes(tmp);
         flb_free(tmp);
         if (limit == -1) {
             return -1;
         }
-        in->mp_buf_limit = (size_t) limit;
+        in->mem_buf_limit = (size_t) limit;
     }
     else if (prop_key_check("listen", k, len) == 0) {
         in->host.listen = tmp;
@@ -261,11 +266,116 @@ int flb_input_set_property(struct flb_input_instance *in, char *k, char *v)
     return 0;
 }
 
-char *flb_input_get_property(char *key, struct flb_input_instance *i)
+const char *flb_input_get_property(const char *key, struct flb_input_instance *i)
 {
     return flb_config_prop_get(key, &i->properties);
 }
 
+/* Return an instance name or alias */
+const char *flb_input_name(struct flb_input_instance *in)
+{
+    if (in->alias) {
+        return in->alias;
+    }
+
+    return in->name;
+}
+
+void flb_input_instance_free(struct flb_input_instance *in)
+{
+    struct mk_list *head_prop;
+    struct flb_config_prop *prop;
+    struct mk_list *tmp_prop;
+
+    if (in->alias) {
+        flb_free(in->alias);
+    }
+
+    /* Remove URI context */
+    if (in->host.uri) {
+        flb_uri_destroy(in->host.uri);
+    }
+    flb_free(in->host.name);
+    flb_free(in->host.address);
+
+    /* release the tag if any */
+    flb_free(in->tag);
+
+    /* Let the engine remove any pending task */
+    flb_engine_destroy_tasks(&in->tasks);
+
+    /* release properties */
+    mk_list_foreach_safe(head_prop, tmp_prop, &in->properties) {
+        prop = mk_list_entry(head_prop, struct flb_config_prop, _head);
+
+        flb_free(prop->key);
+        flb_free(prop->val);
+
+        mk_list_del(&prop->_head);
+        flb_free(prop);
+    }
+
+    /* Remove metrics */
+#ifdef FLB_HAVE_METRICS
+    if (in->metrics) {
+        flb_metrics_destroy(in->metrics);
+    }
+#endif
+
+    if (in->storage) {
+        flb_storage_input_destroy(in);
+    }
+
+    /* Unlink and release */
+    mk_list_del(&in->_head);
+    flb_free(in);
+}
+
+int flb_input_instance_init(struct flb_input_instance *in,
+                            struct flb_config *config)
+{
+    int ret;
+    const char *name;
+    struct flb_input_plugin *p = in->p;
+
+    /* Skip pseudo input plugins */
+    if (!p) {
+        return 0;
+    }
+
+    /* Metrics */
+#ifdef FLB_HAVE_METRICS
+    /* Get name or alias for the instance */
+    name = flb_input_name(in);
+
+    /* Create the metrics context */
+    in->metrics = flb_metrics_create(name);
+    if (in->metrics) {
+        flb_metrics_add(FLB_METRIC_N_RECORDS, "records", in->metrics);
+        flb_metrics_add(FLB_METRIC_N_BYTES, "bytes", in->metrics);
+    }
+#endif
+
+    /* Initialize the input */
+    if (p->cb_init) {
+        /* Sanity check: all non-dynamic tag input plugins must have a tag */
+        if (!in->tag) {
+            flb_input_set_property(in, "tag", in->name);
+        }
+
+        ret = p->cb_init(in, config, in->data);
+        if (ret != 0) {
+            flb_error("Failed initialize input %s",
+                      in->name);
+            flb_input_instance_free(in);
+            return -1;
+        }
+    }
+
+    return 0;
+}
+
+
 /* Initialize all inputs */
 void flb_input_initialize_all(struct flb_config *config)
 {
@@ -288,26 +398,10 @@ void flb_input_initialize_all(struct flb_config *config)
             continue;
         }
 
-        /* Initialize the input */
-        if (p->cb_init) {
-            /* Sanity check: all non-dynamic tag input plugins must have a tag */
-            if (!in->tag && ((p->flags & FLB_INPUT_DYN_TAG) == 0)) {
-                flb_input_set_property(in, "tag", in->name);
-            }
-
-            ret = p->cb_init(in, config, in->data);
-            if (ret != 0) {
-                flb_error("Failed initialize input %s",
-                          in->name);
-                mk_list_del(&in->_head);
-                if (p->flags & FLB_INPUT_NET) {
-                    flb_free(in->tag);
-                    flb_free(in->host.uri);
-                    flb_free(in->host.name);
-                    flb_free(in->host.address);
-                }
-                flb_free(in);
-            }
+        /* Initialize instance */
+        ret = flb_input_instance_init(in, config);
+        if (ret == -1) {
+            /* do nothing, it's ok if it fails */
         }
     }
 }
@@ -332,14 +426,22 @@ void flb_input_pre_run_all(struct flb_config *config)
     }
 }
 
+void flb_input_instance_exit(struct flb_input_instance *in,
+                             struct flb_config *config)
+{
+    struct flb_input_plugin *p;
+
+    p = in->p;
+    if (p->cb_exit && in->context) {
+        p->cb_exit(in->context, config);
+    }
+}
+
 /* Invoke all exit input callbacks */
 void flb_input_exit_all(struct flb_config *config)
 {
     struct mk_list *tmp;
     struct mk_list *head;
-    struct mk_list *tmp_prop;
-    struct mk_list *head_prop;
-    struct flb_config_prop *prop;
     struct flb_input_instance *in;
     struct flb_input_plugin *p;
 
@@ -351,50 +453,8 @@ void flb_input_exit_all(struct flb_config *config)
             continue;
         }
 
-        if (p->cb_exit) {
-            p->cb_exit(in->context, config);
-        }
-
-        /* Remove URI context */
-        if (in->host.uri) {
-            flb_uri_destroy(in->host.uri);
-        }
-        flb_free(in->host.name);
-        flb_free(in->host.address);
-
-        /* Destroy buffer */
-        msgpack_sbuffer_destroy(&in->mp_sbuf);
-        msgpack_zone_free(in->mp_zone);
-
-        /* release the tag if any */
-        flb_free(in->tag);
-
-        /* Let the engine remove any pending task */
-        flb_engine_destroy_tasks(&in->tasks);
-
-        /* release properties */
-        mk_list_foreach_safe(head_prop, tmp_prop, &in->properties) {
-            prop = mk_list_entry(head_prop, struct flb_config_prop, _head);
-
-            flb_free(prop->key);
-            flb_free(prop->val);
-
-            mk_list_del(&prop->_head);
-            flb_free(prop);
-        }
-
-        flb_input_dyntag_exit(in);
-
-        /* Remove metrics */
-#ifdef FLB_HAVE_METRICS
-        if (in->metrics) {
-            flb_metrics_destroy(in->metrics);
-        }
-#endif
-
-        /* Unlink and release */
-        mk_list_del(&in->_head);
-        flb_free(in);
+        flb_input_instance_exit(in, config);
+        flb_input_instance_free(in);
     }
 }
 
@@ -411,6 +471,7 @@ int flb_input_check(struct flb_config *config)
 /*
  * API for Input plugins
  * =====================
+ *  Copyright (C) 2019      The Fluent Bit Authors
  * The Input interface provides a certain number of functions that can be
  * used by Input plugins to configure it own behavior and request specific
  *
@@ -474,7 +535,7 @@ int flb_input_set_collector_time(struct flb_input_instance *in,
     collector->nanoseconds = nanoseconds;
     collector->instance    = in;
     collector->running     = FLB_FALSE;
-    MK_EVENT_NEW(&collector->event);
+    MK_EVENT_ZERO(&collector->event);
     mk_list_add(&collector->_head, &config->collectors);
     mk_list_add(&collector->_head_ins, &in->collectors);
 
@@ -499,7 +560,7 @@ int flb_input_set_collector_event(struct flb_input_instance *in,
     collector->nanoseconds = -1;
     collector->instance    = in;
     collector->running     = FLB_FALSE;
-    MK_EVENT_NEW(&collector->event);
+    MK_EVENT_ZERO(&collector->event);
     mk_list_add(&collector->_head, &config->collectors);
     mk_list_add(&collector->_head_ins, &in->collectors);
 
@@ -544,7 +605,7 @@ static int collector_start(struct flb_input_collector *coll,
                            MK_EVENT_READ, event);
         if (ret == -1) {
             flb_error("[input collector] COLLECT_EVENT registration failed");
-            close(coll->fd_event);
+            mk_event_closesocket(coll->fd_event);
             coll->running = FLB_FALSE;
             return -1;
         }
@@ -634,7 +695,7 @@ int flb_input_pause_all(struct flb_config *config)
             }
             paused++;
         }
-        in->mp_buf_status = FLB_INPUT_PAUSED;
+        in->mem_buf_status = FLB_INPUT_PAUSED;
     }
 
     return paused;
@@ -659,7 +720,7 @@ int flb_input_collector_pause(int coll_id, struct flb_input_instance *in)
          * one can be created.
          */
         mk_event_timeout_destroy(config->evl, &coll->event);
-        close(coll->fd_timer);
+        mk_event_closesocket(coll->fd_timer);
         coll->fd_timer = -1;
     }
     else if (coll->type & (FLB_COLLECT_FD_SERVER | FLB_COLLECT_FD_EVENT)) {
@@ -746,216 +807,13 @@ int flb_input_set_collector_socket(struct flb_input_instance *in,
     collector->nanoseconds = -1;
     collector->instance    = in;
     collector->running     = FLB_FALSE;
-    MK_EVENT_NEW(&collector->event);
+    MK_EVENT_ZERO(&collector->event);
     mk_list_add(&collector->_head, &config->collectors);
     mk_list_add(&collector->_head_ins, &in->collectors);
 
     return 0;
 }
 
-/* Creates a new dyntag node for the input_instance in question */
-struct flb_input_dyntag *flb_input_dyntag_create(struct flb_input_instance *in,
-                                                 char *tag, int tag_len)
-{
-    struct flb_input_dyntag *dt;
-
-    if (tag_len < 1) {
-        return NULL;
-    }
-
-    /* Allocate node and reset fields */
-    dt = flb_malloc(sizeof(struct flb_input_dyntag));
-    if (!dt) {
-        return NULL;
-    }
-    dt->busy = FLB_FALSE;
-    dt->lock = FLB_FALSE;
-    dt->in   = in;
-    dt->tag  = flb_malloc(tag_len + 1);
-    memcpy(dt->tag, tag, tag_len);
-    dt->tag[tag_len] = '\0';
-    dt->tag_len = tag_len;
-
-    /* Initialize MessagePack fields */
-    msgpack_sbuffer_init(&dt->mp_sbuf);
-    msgpack_packer_init(&dt->mp_pck, &dt->mp_sbuf, msgpack_sbuffer_write);
-
-    /* Link to the list head */
-    mk_list_add(&dt->_head, &in->dyntags);
-    return dt;
-}
-
-/* Destroy an dyntag node */
-int flb_input_dyntag_destroy(struct flb_input_dyntag *dt)
-{
-    flb_debug("[dyntag %s] %p destroy (tag=%s, bytes=%lu)",
-              dt->in->name, dt, dt->tag, dt->mp_sbuf.size);
-
-    msgpack_sbuffer_destroy(&dt->mp_sbuf);
-    mk_list_del(&dt->_head);
-    flb_free(dt->tag);
-    flb_free(dt);
-
-    return 0;
-}
-
-void flb_input_dyntag_exit(struct flb_input_instance *in)
-{
-    struct mk_list *tmp;
-    struct mk_list *head;
-    struct flb_input_dyntag *dt;
-
-    mk_list_foreach_safe(head, tmp, &in->dyntags) {
-        dt = mk_list_entry(head, struct flb_input_dyntag, _head);
-        flb_input_dyntag_destroy(dt);
-    }
-}
-
-struct flb_input_dyntag *flb_input_dyntag_get(char *tag, size_t tag_len,
-                                              struct flb_input_instance *in)
-
-{
-    struct mk_list *head;
-    struct flb_input_dyntag *dt = NULL;
-
-    /* Try to find a current dyntag node to append the data */
-    mk_list_foreach(head, &in->dyntags) {
-        dt = mk_list_entry(head, struct flb_input_dyntag, _head);
-        if (dt->busy == FLB_TRUE || dt->lock == FLB_TRUE) {
-            dt = NULL;
-            continue;
-        }
-
-        if (dt->tag_len != tag_len) {
-            dt = NULL;
-            continue;
-        }
-
-        if (strncmp(dt->tag, tag, tag_len) != 0) {
-            dt = NULL;
-            continue;
-        }
-        break;
-    }
-
-    /* No dyntag was found, we need to create a new one */
-    if (!dt) {
-        dt = flb_input_dyntag_create(in, tag, tag_len);
-        if (!dt) {
-            return NULL;
-        }
-    }
-
-    return dt;
-}
-
-/* Append a MessagPack Object to the input instance */
-int flb_input_dyntag_append_obj(struct flb_input_instance *in,
-                                char *tag, size_t tag_len,
-                                msgpack_object data)
-{
-    struct flb_input_dyntag *dt;
-
-    dt = flb_input_dyntag_get(tag, tag_len, in);
-    if (!dt) {
-        return -1;
-    }
-
-    flb_input_dbuf_write_start(dt);
-    msgpack_pack_object(&dt->mp_pck, data);
-    flb_input_dbuf_write_end(dt);
-
-    /* Lock buffers where size > 2MB */
-    if (dt->mp_sbuf.size > 2048000) {
-        dt->lock = FLB_TRUE;
-    }
-
-    return 0;
-}
-
-/* Append a RAW MessagPack buffer to the input instance */
-int flb_input_dyntag_append_raw(struct flb_input_instance *in,
-                                char *tag, size_t tag_len,
-                                void *buf, size_t buf_size)
-{
-    struct flb_input_dyntag *dt;
-
-    dt = flb_input_dyntag_get(tag, tag_len, in);
-    if (!dt) {
-        return -1;
-    }
-
-    /* Mark buf write */
-    flb_input_dbuf_write_start(dt);
-
-    msgpack_sbuffer_write(&dt->mp_sbuf, buf, buf_size);
-
-    /* Unmark buf write */
-    flb_input_dbuf_write_end(dt);
-
-    /* Lock buffers where size > 2MB */
-    if (dt->mp_sbuf.size > 2048000) {
-        dt->lock = FLB_TRUE;
-    }
-
-    return 0;
-}
-
-/* Flush a buffer from an input instance (new since v0.11) */
-void *flb_input_flush(struct flb_input_instance *i_ins, size_t *size)
-{
-    char *buf;
-
-    if (i_ins->mp_sbuf.size == 0) {
-        *size = 0;
-        return NULL;
-    }
-
-    /* Allocate buffer */
-    buf = flb_malloc(i_ins->mp_sbuf.size);
-    if (!buf) {
-        flb_errno();
-        return NULL;
-    }
-
-    /* Copy original data to new buffer and update it size */
-    memcpy(buf, i_ins->mp_sbuf.data, i_ins->mp_sbuf.size);
-    *size = i_ins->mp_sbuf.size;
-
-    /* re-initialize msgpack buffers */
-    i_ins->mp_records = 0;
-    msgpack_sbuffer_destroy(&i_ins->mp_sbuf);
-    msgpack_sbuffer_init(&i_ins->mp_sbuf);
-
-    return buf;
-}
-
-/* Retrieve a raw buffer from a dyntag node */
-void *flb_input_dyntag_flush(struct flb_input_dyntag *dt, size_t *size)
-{
-    void *buf;
-
-    /*
-     * msgpack-c internal use a raw buffer for it operations, since we
-     * already appended data we just can take out the references to avoid
-     * a new memory allocation and skip a copy operation.
-     */
-
-    buf   = dt->mp_sbuf.data;
-    *size = dt->mp_sbuf.size;
-
-    /* Unset the lock, it means more data can be added */
-    //dt->lock = FLB_FALSE;
-
-    /* Set it busy as it likely it's a reference for an outgoing task */
-    dt->busy = FLB_TRUE;
-
-    //msgpack_sbuffer_init(&dt->mp_sbuf);
-    //msgpack_packer_init(&dt->mp_pck, &dt->mp_sbuf, msgpack_sbuffer_write);
-
-    return buf;
-}
-
 int flb_input_collector_fd(flb_pipefd_t fd, struct flb_config *config)
 {
     struct mk_list *head;
