@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -26,10 +27,11 @@
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_error.h>
 #include <fluent-bit/flb_utils.h>
+#include <fluent-bit/flb_config.h>
+#include <fluent-bit/flb_strptime.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <time.h>
 #include <limits.h>
 #include <string.h>
 
@@ -82,19 +84,29 @@ static unsigned u64_to_str(uint64_t value, char* dst) {
 }
 
 int flb_parser_regex_do(struct flb_parser *parser,
-                        char *buf, size_t length,
+                        const char *buf, size_t length,
                         void **out_buf, size_t *out_size,
                         struct flb_time *out_time);
 
 int flb_parser_json_do(struct flb_parser *parser,
-                       char *buf, size_t length,
+                       const char *buf, size_t length,
                        void **out_buf, size_t *out_size,
                        struct flb_time *out_time);
 
-struct flb_parser *flb_parser_create(char *name, char *format,
-                                     char *p_regex,
-                                     char *time_fmt, char *time_key,
-                                     char *time_offset,
+int flb_parser_ltsv_do(struct flb_parser *parser,
+                       const char *buf, size_t length,
+                       void **out_buf, size_t *out_size,
+                       struct flb_time *out_time);
+
+int flb_parser_logfmt_do(struct flb_parser *parser,
+                         const char *buf, size_t length,
+                         void **out_buf, size_t *out_size,
+                         struct flb_time *out_time);
+
+struct flb_parser *flb_parser_create(const char *name, const char *format,
+                                     const char *p_regex,
+                                     const char *time_fmt, const char *time_key,
+                                     const char *time_offset,
                                      int time_keep,
                                      struct flb_parser_types *types,
                                      int types_len,
@@ -129,12 +141,18 @@ struct flb_parser *flb_parser_create(char *name, char *format,
     p->decoders = decoders;
 
     /* Format lookup */
-    if (strcmp(format, "regex") == 0) {
+    if (strcasecmp(format, "regex") == 0) {
         p->type = FLB_PARSER_REGEX;
     }
-    else if (strcmp(format, "json") == 0) {
+    else if (strcasecmp(format, "json") == 0) {
         p->type = FLB_PARSER_JSON;
     }
+    else if (strcmp(format, "ltsv") == 0) {
+        p->type = FLB_PARSER_LTSV;
+    }
+    else if (strcmp(format, "logfmt") == 0) {
+        p->type = FLB_PARSER_LOGFMT;
+    }
     else {
         flb_error("[parser:%s] Invalid format %s", name, format);
         flb_free(p);
@@ -148,7 +166,7 @@ struct flb_parser *flb_parser_create(char *name, char *format,
             return NULL;
         }
 
-        regex = flb_regex_create((unsigned char *) p_regex);
+        regex = flb_regex_create(p_regex);
         if (!regex) {
             flb_error("[parser:%s] Invalid regex pattern %s", name, p_regex);
             flb_free(p);
@@ -188,6 +206,19 @@ struct flb_parser *flb_parser_create(char *name, char *format,
             *tmp++ = '\0';
         }
 
+        /* Check if the format contains a timezone (%z) */
+        if (strstr(p->time_fmt, "%z") || strstr(p->time_fmt, "%Z") ||
+            strstr(p->time_fmt, "%SZ") || strstr(p->time_fmt, "%S.%LZ")) {
+#if defined(FLB_HAVE_GMTOFF) || !defined(FLB_HAVE_SYSTEM_STRPTIME)
+            p->time_with_tz = FLB_TRUE;
+#else
+            flb_error("[parser] timezone offset not supported");
+            flb_error("[parser] you cannot use %%z/%%Z on this platform");
+            flb_free(p);
+            return NULL;
+#endif
+        }
+
         /*
          * Check if the format expect fractional seconds
          *
@@ -212,19 +243,27 @@ struct flb_parser *flb_parser_create(char *name, char *format,
         }
         if (tmp) {
             tmp[2] = '\0';
-            p->time_frac_secs = (tmp + 3);
+            p->time_frac_secs = (tmp + 5);
         }
         else {
-            p->time_frac_secs = NULL;
-        }
+            /* same as above but with comma seperator */
+            if (p->time_with_year == FLB_TRUE) {
+                tmp = strstr(p->time_fmt, "%S,%L");
+            }
+            else {
+                tmp = strstr(p->time_fmt_year, "%s,%L");
 
-        /* Check if the format contains a timezone (%z) */
-        tmp = strstr(p->time_fmt, "%z");
-        if (tmp) {
-            p->time_with_tz = FLB_TRUE;
-        }
-        else if (strstr(p->time_fmt, "%SZ") || strstr(p->time_fmt, "%S.%LZ")) {
-            p->time_with_tz = FLB_TRUE;
+                if (tmp == NULL) {
+                    tmp = strstr(p->time_fmt_year, "%S,%L");
+                }
+            }
+            if (tmp) {
+                tmp[2] = '\0';
+                p->time_frac_secs = (tmp + 5);
+            }
+            else {
+                p->time_frac_secs = NULL;
+            }
         }
 
         /* Optional fixed timezone offset */
@@ -298,11 +337,12 @@ void flb_parser_exit(struct flb_config *config)
     }
 }
 
-static int proc_types_str(char *types_str, struct flb_parser_types **types)
+static int proc_types_str(const char *types_str, struct flb_parser_types **types)
 {
     int i = 0;
     int types_num = 0;
     char *type_str = NULL;
+    size_t len;
     struct mk_list *split;
     struct mk_list *head;
     struct flb_split_entry *sentry;
@@ -324,10 +364,9 @@ static int proc_types_str(char *types_str, struct flb_parser_types **types)
             i++;
             continue;
         }
-        *type_str = '\0'; /* for strdup */
-        (*types)[i].key = flb_strdup(sentry->value);
-        (*types)[i].key_len = strlen(sentry->value);
-        *types_str = ':';
+        len = type_str - sentry->value;
+        (*types)[i].key = flb_strndup(sentry->value, len);
+        (*types)[i].key_len = len;
 
         type_str++;
         if (!strcasecmp(type_str, "integer")) {
@@ -353,11 +392,11 @@ static int proc_types_str(char *types_str, struct flb_parser_types **types)
 }
 
 /* Load parsers from a configuration file */
-int flb_parser_conf_file(char *file, struct flb_config *config)
+int flb_parser_conf_file(const char *file, struct flb_config *config)
 {
     int ret;
     char tmp[PATH_MAX + 1];
-    char *cfg = NULL;
+    const char *cfg = NULL;
     char *name;
     char *format;
     char *regex;
@@ -372,9 +411,10 @@ int flb_parser_conf_file(char *file, struct flb_config *config)
     struct mk_rconf_section *section;
     struct mk_list *head;
     struct stat st;
-    struct flb_parser_types *types;
+    struct flb_parser_types *types = NULL;
     struct mk_list *decoders;
 
+#ifndef FLB_HAVE_STATIC_CONF
     ret = stat(file, &st);
     if (ret == -1 && errno == ENOENT) {
         /* Try to resolve the real path (if exists) */
@@ -392,8 +432,11 @@ int flb_parser_conf_file(char *file, struct flb_config *config)
         cfg = file;
     }
 
-    flb_debug("[parser] opening file %s", cfg);
     fconf = mk_rconf_open(cfg);
+#else
+    fconf = flb_config_static_open(file);
+#endif
+
     if (!fconf) {
         return -1;
     }
@@ -416,21 +459,21 @@ int flb_parser_conf_file(char *file, struct flb_config *config)
         /* Name */
         name = mk_rconf_section_get_key(section, "Name", MK_RCONF_STR);
         if (!name) {
-            flb_error("[parser] no parser 'name' found");
+            flb_error("[parser] no parser 'name' found in file '%s'", cfg);
             goto fconf_error;
         }
 
         /* Format */
         format = mk_rconf_section_get_key(section, "Format", MK_RCONF_STR);
         if (!format) {
-            flb_error("[parser] no parser 'format' found");
+            flb_error("[parser] no parser 'format' found for '%s' in file '%s'", name, cfg);
             goto fconf_error;
         }
 
         /* Regex (if format is regex) */
         regex = mk_rconf_section_get_key(section, "Regex", MK_RCONF_STR);
         if (!regex && strcmp(format, "regex") == 0) {
-            flb_error("[parser] no parser 'regex' found");
+            flb_error("[parser] no parser 'regex' found for '%s' in file '%s", name, cfg);
             goto fconf_error;
         }
 
@@ -526,7 +569,7 @@ int flb_parser_conf_file(char *file, struct flb_config *config)
     return -1;
 }
 
-struct flb_parser *flb_parser_get(char *name, struct flb_config *config)
+struct flb_parser *flb_parser_get(const char *name, struct flb_config *config)
 {
     struct mk_list *head;
     struct flb_parser *parser;
@@ -542,7 +585,7 @@ struct flb_parser *flb_parser_get(char *name, struct flb_config *config)
     return NULL;
 }
 
-int flb_parser_do(struct flb_parser *parser, char *buf, size_t length,
+int flb_parser_do(struct flb_parser *parser, const char *buf, size_t length,
                   void **out_buf, size_t *out_size, struct flb_time *out_time)
 {
 
@@ -554,18 +597,26 @@ int flb_parser_do(struct flb_parser *parser, char *buf, size_t length,
         return flb_parser_json_do(parser, buf, length,
                                   out_buf, out_size, out_time);
     }
+    else if (parser->type == FLB_PARSER_LTSV) {
+        return flb_parser_ltsv_do(parser, buf, length,
+                                  out_buf, out_size, out_time);
+    }
+    else if (parser->type == FLB_PARSER_LOGFMT) {
+        return flb_parser_logfmt_do(parser, buf, length,
+                                  out_buf, out_size, out_time);
+    }
 
     return -1;
 }
 
 /* Given a timezone string, return it numeric offset */
-int flb_parser_tzone_offset(char *str, int len, int *tmdiff)
+int flb_parser_tzone_offset(const char *str, int len, int *tmdiff)
 {
     int neg;
     long hour;
     long min;
-    char *end;
-    char *p = str;
+    const char *end;
+    const char *p = str;
 
     /* Check timezones */
     if (*p == 'Z') {
@@ -607,20 +658,19 @@ int flb_parser_tzone_offset(char *str, int len, int *tmdiff)
     return 0;
 }
 
-int flb_parser_time_lookup(char *time_str, size_t tsize,
+int flb_parser_time_lookup(const char *time_str, size_t tsize,
                            time_t now,
                            struct flb_parser *parser,
                            struct tm *tm, double *ns)
 {
     int ret;
     int slen;
-    int tmdiff = 0;
     time_t time_now;
     double tmfrac = 0;
     char *p = NULL;
     char *fmt;
     int time_len = tsize;
-    char *time_ptr = time_str;
+    const char *time_ptr = time_str;
     char tmp[64];
     char fs_tmp[32];
     struct tm tmy;
@@ -676,7 +726,7 @@ int flb_parser_time_lookup(char *time_str, size_t tsize,
 
     if (p != NULL) {
         /* Check if we have fractional seconds */
-        if (parser->time_frac_secs && *p == '.') {
+        if (parser->time_frac_secs && (*p == '.' || *p == ',')) {
             /*
              * Further parser routines needs a null byte, for fractional seconds
              * we make a safe copy of the content.
@@ -689,73 +739,76 @@ int flb_parser_time_lookup(char *time_str, size_t tsize,
             fs_tmp[slen] = '\0';
 
             /* Parse fractional seconds */
-            tmdiff = 0;
-            ret = flb_parser_frac_tzone(fs_tmp, slen, &tmfrac, &tmdiff);
+            ret = flb_parser_frac(fs_tmp, slen, &tmfrac, &time_ptr);
             if (ret == -1) {
-                /* Check if the timezone failed */
-                if (tmdiff == -1) {
-                    /* Try to find a workaround for time offset resolution */
-                    tm->tm_gmtoff = parser->time_offset;
-                }
-                else {
-                    flb_warn("[parser] Error parsing time string");
-                    return -1;
-                }
-            }
-            else {
-                tm->tm_gmtoff = tmdiff;
+                flb_warn("[parser] Error parsing time string");
+                return -1;
             }
             *ns = tmfrac;
-        }
-        else {
-            if (parser->time_with_tz == FLB_FALSE) {
-                tm->tm_gmtoff = parser->time_offset;
+
+            p = strptime(time_ptr, parser->time_frac_secs, tm);
+
+            if (p == NULL) {
+                return -1;
             }
         }
+
+#ifdef FLB_HAVE_GMTOFF
+        if (parser->time_with_tz == FLB_FALSE) {
+            tm->tm_gmtoff = parser->time_offset;
+        }
+#endif
+
         return 0;
     }
 
     return -1;
 }
 
-int flb_parser_frac_tzone(char *str, int len, double *frac, int *tmdiff)
+int flb_parser_frac(const char *str, int len, double *frac, const char **end)
 {
-    int ret;
-    int tz_len;
+    int ret = 0;
     char *p;
-    char *end;
     double d;
+    const char *pstr;
+    char *tmp = NULL;
 
     /* Fractional seconds */
-    d = strtod(str, &end);
-    if ((d == 0 && end == str) || !end) {
-        return -1;
+    /* Normalize the fractional seperator to be '.' since that's what strtod()
+     * expects in standard C locale */
+    if (*str == ',') {
+        tmp = flb_strdup(str);
+        tmp[0] = '.';
+        pstr = tmp;
+    }
+    else {
+        pstr = str;
+    }
+
+    d = strtod(pstr, &p);
+    if ((d == 0 && p == pstr) || !p) {
+        ret = -1;
+        goto free_and_return;
     }
     *frac = d;
+    *end = str + (p - pstr);
 
-    tz_len = len - (end - str);
-    p = end;
-    while (*p == ' ') p++;
-
-    tz_len -= (p - end);
-    ret = flb_parser_tzone_offset(p, tz_len, tmdiff);
-    if (ret == -1) {
-        *tmdiff = -1;
-        return -1;
+free_and_return:
+    if (tmp != NULL) {
+        flb_free(tmp);
     }
-
-    return 0;
+    return ret;
 }
 
-int flb_parser_typecast(char *key, int key_len,
-                        char *val, int val_len,
+int flb_parser_typecast(const char *key, int key_len,
+                        const char *val, int val_len,
                         msgpack_packer *pck,
                         struct flb_parser_types *types,
                         int types_len)
 {
     int i;
     int error = FLB_FALSE;
-    char tmp_char;
+    char *tmp_str;
     int casted = FLB_FALSE;
 
     for(i=0; i<types_len; i++){
@@ -774,24 +827,20 @@ int flb_parser_typecast(char *key, int key_len,
                     long long lval;
 
                     /* msgpack char is not null terminated.
-                       So backup and fill null char,
-                       convert int,
-                       rewind char.
+                       So make a temporary copy.
                      */
-                    tmp_char = val[val_len];
-                    val[val_len] = '\0';
-                    lval = atoll(val);
-                    val[val_len] = tmp_char;
+                    tmp_str = flb_strndup(val, val_len);
+                    lval = atoll(tmp_str);
+                    flb_free(tmp_str);
                     msgpack_pack_int64(pck, lval);
                 }
                 break;
             case FLB_PARSER_TYPE_HEX:
                 {
                     unsigned long long lval;
-                    tmp_char = val[val_len];
-                    val[val_len] = '\0';
-                    lval = strtoull(val, NULL, 16);
-                    val[val_len] = tmp_char;
+                    tmp_str = flb_strndup(val, val_len);
+                    lval = strtoull(tmp_str, NULL, 16);
+                    flb_free(tmp_str);
                     msgpack_pack_uint64(pck, lval);
                 }
                 break;
@@ -799,10 +848,9 @@ int flb_parser_typecast(char *key, int key_len,
             case FLB_PARSER_TYPE_FLOAT:
                 {
                     double dval;
-                    tmp_char = val[val_len];
-                    val[val_len] = '\0';
-                    dval = atof(val);
-                    val[val_len] = tmp_char;
+                    tmp_str = flb_strndup(val, val_len);
+                    dval = atof(tmp_str);
+                    flb_free(tmp_str);
                     msgpack_pack_double(pck, dval);
                 }
                 break;
