@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -25,9 +26,15 @@
 #include <fluent-bit/flb_parser_decoder.h>
 #include <fluent-bit/flb_regex.h>
 #include <fluent-bit/flb_str.h>
+
 #include <msgpack.h>
 
+/* don't do this at home */
+#define pack_uint16(buf, d) _msgpack_store16(buf, (uint16_t) d)
+#define pack_uint32(buf, d) _msgpack_store32(buf, (uint32_t) d)
+
 struct regex_cb_ctx {
+    int num_skipped;
     time_t time_lookup;
     time_t time_now;
     double time_frac;
@@ -35,19 +42,25 @@ struct regex_cb_ctx {
     msgpack_packer *pck;
 };
 
-static void cb_results(unsigned char *name, unsigned char *value,
+static void cb_results(const char *name, const char *value,
                        size_t vlen, void *data)
 {
     int len;
     int ret;
     double frac = 0;
     char *time_key;
+    char tmp[255];
     struct regex_cb_ctx *pcb = data;
     struct flb_parser *parser = pcb->parser;
     struct tm tm = {0};
     (void) data;
 
-    len = strlen((char *) name);
+    if (vlen == 0) {
+        pcb->num_skipped++;
+        return;
+    }
+
+    len = strlen(name);
 
     /* Check if there is a time lookup field */
     if (parser->time_fmt) {
@@ -58,40 +71,49 @@ static void cb_results(unsigned char *name, unsigned char *value,
             time_key = "time";
         }
 
-        if (strcmp((char *) name, time_key) == 0) {
+        if (strcmp(name, time_key) == 0) {
             /* Lookup time */
-            ret = flb_parser_time_lookup((char *) value, vlen,
+            ret = flb_parser_time_lookup(value, vlen,
                                          pcb->time_now, parser, &tm, &frac);
             if (ret == -1) {
-                flb_error("[parser:%s] Invalid time format %s.", parser->name, parser->time_fmt);
+                if (vlen > sizeof(tmp) - 1) {
+                    vlen = sizeof(tmp) - 1;
+                }
+                memcpy(tmp, value, vlen);
+                tmp[vlen] = '\0';
+                flb_warn("[parser:%s] Invalid time format %s for '%s'.",
+                         parser->name, parser->time_fmt, tmp);
+                pcb->num_skipped++;
                 return;
             }
+
             pcb->time_frac = frac;
             pcb->time_lookup = flb_parser_tm2time(&tm);
 
             if (parser->time_keep == FLB_FALSE) {
+                pcb->num_skipped++;
                 return;
             }
         }
     }
 
     if (parser->types_len != 0) {
-        flb_parser_typecast((char*)name, len,
-                            (char*)value, vlen,
+        flb_parser_typecast(name, len,
+                            value, vlen,
                             pcb->pck,
                             parser->types,
                             parser->types_len);
     }
     else {
         msgpack_pack_str(pcb->pck, len);
-        msgpack_pack_str_body(pcb->pck, (char *) name, len);
+        msgpack_pack_str_body(pcb->pck, name, len);
         msgpack_pack_str(pcb->pck, vlen);
-        msgpack_pack_str_body(pcb->pck, (char *) value, vlen);
+        msgpack_pack_str_body(pcb->pck, value, vlen);
     }
 }
 
 int flb_parser_regex_do(struct flb_parser *parser,
-                        char *buf, size_t length,
+                        const char *buf, size_t length,
                         void **out_buf, size_t *out_size,
                         struct flb_time *out_time)
 {
@@ -101,13 +123,14 @@ int flb_parser_regex_do(struct flb_parser *parser,
     ssize_t n;
     size_t dec_out_size;
     char *dec_out_buf;
+    char *tmp;
     struct flb_regex_search result;
     struct regex_cb_ctx pcb;
     struct flb_time *t;
     msgpack_sbuffer tmp_sbuf;
     msgpack_packer tmp_pck;
 
-    n = flb_regex_do(parser->regex, (unsigned char *) buf, length, &result);
+    n = flb_regex_do(parser->regex, buf, length, &result);
     if (n <= 0) {
         return -1;
     }
@@ -116,18 +139,14 @@ int flb_parser_regex_do(struct flb_parser *parser,
     msgpack_sbuffer_init(&tmp_sbuf);
     msgpack_packer_init(&tmp_pck, &tmp_sbuf, msgpack_sbuffer_write);
 
-    if (parser->time_fmt && parser->time_keep == FLB_FALSE) {
-        arr_size = (n - 1);
-    }
-    else {
-        arr_size = n;
-    }
-
+    /* Set a Map size with the exact number of matches returned by regex */
+    arr_size = n;
     msgpack_pack_map(&tmp_pck, arr_size);
 
     /* Callback context */
     pcb.pck = &tmp_pck;
     pcb.parser = parser;
+    pcb.num_skipped = 0;
     pcb.time_lookup = 0;
     pcb.time_frac = 0;
     pcb.time_now = time(NULL);
@@ -139,6 +158,47 @@ int flb_parser_regex_do(struct flb_parser *parser,
         return -1;
     }
 
+    /*
+     * There some special cases when the Parser have a 'time' handling
+     * requirement, meaning: lookup for this 'time' key and resolve the
+     * real date of the record. If so, the parser by default will
+     * keep the original 'time' key field found but in other scenarios
+     * it may ask to skip it.
+     *
+     * If a time lookup is specified and the parser ask to skip the record
+     * and the time key is found, we need to adjust the msgpack header
+     * map size, initially we set a size to include all keys found, but
+     * until now we just know we are not going to include it.
+     *
+     * In addition, keys without associated values are skipped too and we
+     * must take this into account in msgpack header map size adjustment.
+     *
+     * In order to avoid to create a new msgpack buffer and repack the
+     * map entries, we just position at the header byte and do the
+     * proper adjustment in our original buffer. Note that for cases
+     * where the map is large enough '<= 65535' or '> 65535' we have
+     * to use internal msgpack api functions since packing the bytes
+     * in Big-Endian is a requirement.
+     */
+     if (pcb.num_skipped > 0) {
+
+        arr_size = (n - pcb.num_skipped);
+
+        tmp = tmp_sbuf.data;
+        uint8_t h = tmp[0];
+        if (h >> 4 == 0x8) { /* 1000xxxx */
+            *tmp = (uint8_t) 0x8 << 4 | ((uint8_t) arr_size);
+        }
+        else if (h == 0xde) {
+            tmp++;
+            pack_uint16(tmp, arr_size);
+        }
+        else if (h == 0xdf) {
+            tmp++;
+            pack_uint32(tmp, arr_size);
+        }
+    }
+
     /* Export results */
     *out_buf = tmp_sbuf.data;
     *out_size = tmp_sbuf.size;
