@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -89,6 +90,9 @@ void flb_utils_error(int err)
     case FLB_ERR_JSON_PART:
         msg = "Truncated JSON string";
         break;
+    case FLB_ERR_CORO_STACK_SIZE:
+        msg = "Invalid coroutine stack size";
+        break;
     }
 
     if (!msg) {
@@ -197,7 +201,7 @@ void flb_utils_print_setup(struct flb_config *config)
     }
 }
 
-struct mk_list *flb_utils_split(char *line, int separator, int max_split)
+struct mk_list *flb_utils_split(const char *line, int separator, int max_split)
 {
     int i = 0;
     int count = 0;
@@ -208,6 +212,10 @@ struct mk_list *flb_utils_split(char *line, int separator, int max_split)
     struct mk_list *list;
     struct flb_split_entry *new;
 
+    if (!line) {
+        return NULL;
+    }
+
     list = flb_malloc(sizeof(struct mk_list));
     if (!list) {
         flb_errno();
@@ -233,20 +241,40 @@ struct mk_list *flb_utils_split(char *line, int separator, int max_split)
             val = mk_string_copy_substr(line, i, len);
             val_len = len - i;
             end = len;
-
         }
 
+        /* Update last position */
+        i = end;
+
         /* Create new entry */
         new = flb_malloc(sizeof(struct flb_split_entry));
+        if (!new) {
+            flb_errno();
+            flb_free(val);
+            flb_utils_split_free(list);
+            return NULL;
+        }
         new->value = val;
         new->len = val_len;
-
+        new->last_pos = i;
         mk_list_add(&new->_head, list);
-        i = end + 1;
-
         count++;
+
+        /* Update index for next loop */
+        i++;
+
+        /*
+         * If the counter exceeded the maximum specified and there
+         * are still remaining bytes, append those bytes in a new
+         * and last entry.
+         */
         if (count >= max_split && max_split > 0 && i < len) {
             new = flb_malloc(sizeof(struct flb_split_entry));
+            if (!new) {
+                flb_errno();
+                flb_utils_split_free(list);
+                return NULL;
+            }
             new->value = mk_string_copy_substr(line, i, len);
             new->len   = len - i;
             mk_list_add(&new->_head, list);
@@ -309,12 +337,12 @@ int flb_utils_pipe_byte_consume(flb_pipefd_t fd)
     return 0;
 }
 
-ssize_t flb_utils_size_to_bytes(char *size)
+int64_t flb_utils_size_to_bytes(const char *size)
 {
     int i;
     int len;
     int plen = 0;
-    size_t val;
+    int64_t val;
     char c;
     char tmp[3] = {0};
     int64_t KB = 1000;
@@ -375,7 +403,7 @@ ssize_t flb_utils_size_to_bytes(char *size)
     return val;
 }
 
-int flb_utils_time_to_seconds(char *time)
+int flb_utils_time_to_seconds(const char *time)
 {
     int len;
     size_t val;
@@ -397,7 +425,7 @@ int flb_utils_time_to_seconds(char *time)
     return val;
 }
 
-int flb_utils_bool(char *val)
+int flb_utils_bool(const char *val)
 {
     if (strcasecmp(val, "true") == 0 ||
         strcasecmp(val, "on") == 0 ||
@@ -409,7 +437,7 @@ int flb_utils_bool(char *val)
 }
 
 /* Convert a 'string' time seconds.nanoseconds to int and long values */
-int flb_utils_time_split(char *time, int *sec, long *nsec)
+int flb_utils_time_split(const char *time, int *sec, long *nsec)
 {
     char *p;
     char *end;
@@ -473,7 +501,7 @@ void flb_utils_bytes_to_human_readable_size(size_t bytes,
 }
 
 
-static inline void encoded_to_buf(char *out, char *in, int len)
+static inline void encoded_to_buf(char *out, const char *in, int len)
 {
     int i;
     char *p = out;
@@ -489,7 +517,7 @@ static inline void encoded_to_buf(char *out, char *in, int len)
  * representation.
  */
 int flb_utils_write_str(char *buf, int *off, size_t size,
-                        char *str, size_t str_len)
+                        const char *str, size_t str_len)
 {
     int i;
     int b;
@@ -512,7 +540,6 @@ int flb_utils_write_str(char *buf, int *off, size_t size,
         return FLB_FALSE;
     }
 
-    written = *off;
     p = buf + *off;
     for (i = 0; i < str_len; i++) {
         if ((available - written) < 2) {
@@ -520,35 +547,33 @@ int flb_utils_write_str(char *buf, int *off, size_t size,
         }
 
         c = (uint32_t) str[i];
-        if (c == '\\' || c == '"') {
+        if (c == '\"') {
             *p++ = '\\';
-            *p++ = c;
+            *p++ = '\"';
         }
-        else if (c >= '\a' && c <= '\r') {
+        else if (c == '\\') {
             *p++ = '\\';
-            switch (c) {
-            case '\n':
-                *p++ = 'n';
-                break;
-            case '\t':
-                *p++ = 't';
-                break;
-            case '\b':
-                *p++ = 'b';
-                break;
-            case '\f':
-                *p++ = 'f';
-                break;
-            case '\r':
-                *p++ = 'r';
-                break;
-            case '\a':
-                *p++ = 'a';
-                break;
-            case '\v':
-                *p++ = 'v';
-                break;
-            }
+            *p++ = '\\';
+        }
+        else if (c == '\n') {
+            *p++ = '\\';
+            *p++ = 'n';
+        }
+        else if (c == '\r') {
+            *p++ = '\\';
+            *p++ = 'r';
+        }
+        else if (c == '\t') {
+            *p++ = '\\';
+            *p++ = 't';
+        }
+        else if (c == '\b') {
+            *p++ = '\\';
+            *p++ = 'b';
+        }
+        else if (c == '\f') {
+            *p++ = '\\';
+            *p++ = 'f';
         }
         else if (c < 32 || c == 0x7f) {
             if ((available - written) < 6) {
@@ -624,7 +649,7 @@ int flb_utils_write_str(char *buf, int *off, size_t size,
 }
 
 
-int flb_utils_write_str_buf(char *str, size_t str_len, char **out, size_t *out_size)
+int flb_utils_write_str_buf(const char *str, size_t str_len, char **out, size_t *out_size)
 {
     int ret;
     int off;
@@ -662,3 +687,89 @@ int flb_utils_write_str_buf(char *str, size_t str_len, char **out, size_t *out_s
     *out_size = off;
     return 0;
 }
+
+int flb_utils_url_split(const char *in_url, char **out_protocol,
+                        char **out_host, char **out_port, char **out_uri)
+{
+    char *protocol = NULL;
+    char *host = NULL;
+    char *port = NULL;
+    char *uri = NULL;
+    char *p;
+    char *tmp;
+    char *sep;
+
+    /* Protocol */
+    p = strstr(in_url, "://");
+    if (!p) {
+        return -1;
+    }
+    if (p == in_url) {
+        return -1;
+    }
+
+    protocol = mk_string_copy_substr(in_url, 0, p - in_url);
+    if (!protocol) {
+        flb_errno();
+        return -1;
+    }
+
+    /* Advance position after protocol */
+    p += 3;
+
+    /* Check for first '/' */
+    sep = strchr(p, '/');
+    tmp = strchr(p, ':');
+
+    /* Validate port separator is found before the first slash */
+    if (sep && tmp) {
+        if (tmp > sep) {
+            tmp = NULL;
+        }
+    }
+
+    if (tmp) {
+        host = mk_string_copy_substr(p, 0, tmp - p);
+        if (!host) {
+            flb_errno();
+            goto error;
+        }
+        p = tmp + 1;
+
+        /* Look for an optional URI */
+        tmp = strchr(p, '/');
+        if (tmp) {
+            port = mk_string_copy_substr(p, 0, tmp - p);
+            uri = flb_strdup(tmp);
+        }
+        else {
+            port = flb_strdup(p);
+            uri = flb_strdup("/");
+        }
+    }
+    else {
+        tmp = strchr(p, '/');
+        if (tmp) {
+            host = mk_string_copy_substr(p, 0, tmp - p);
+            uri = flb_strdup(tmp);
+        }
+        else {
+            host = flb_strdup(p);
+            uri = flb_strdup("/");
+        }
+    }
+
+    *out_protocol = protocol;
+    *out_host = host;
+    *out_port = port;
+    *out_uri = uri;
+
+    return 0;
+
+ error:
+    if (protocol) {
+        flb_free(protocol);
+    }
+
+    return -1;
+}
