@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -26,6 +27,7 @@
 #include <signal.h>
 
 #include <monkey/mk_core.h>
+#include <fluent-bit/flb_compat.h>
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_env.h>
 #include <fluent-bit/flb_macros.h>
@@ -39,7 +41,8 @@
 #include <fluent-bit/flb_filter.h>
 #include <fluent-bit/flb_engine.h>
 #include <fluent-bit/flb_str.h>
-#include <fluent-bit/flb_plugin_proxy.h>
+#include <fluent-bit/flb_slist.h>
+#include <fluent-bit/flb_plugin.h>
 #include <fluent-bit/flb_parser.h>
 
 /* Libbacktrace support */
@@ -121,10 +124,7 @@ static void flb_help(int rc, struct flb_config *config)
 
     printf("Usage: fluent-bit [OPTION]\n\n");
     printf("%sAvailable Options%s\n", ANSI_BOLD, ANSI_RESET);
-#ifdef FLB_HAVE_BUFFERING
-    printf("  -b  --buf_path=PATH\tspecify a buffering path\n");
-    printf("  -B  --buf_workers=N\tnumber of workers for buffering\n");
-#endif
+    printf("  -b  --storage_path=PATH\tspecify a storage buffering path\n");
     printf("  -c  --config=FILE\tspecify an optional configuration file\n");
 #ifdef FLB_HAVE_FORK
     printf("  -d, --daemon\t\trun Fluent Bit in background mode\n");
@@ -136,16 +136,23 @@ static void flb_help(int rc, struct flb_config *config)
     printf("  -m, --match=MATCH\tset plugin match, same as '-p match=abc'\n");
     printf("  -o, --output=OUTPUT\tset an output\n");
     printf("  -p, --prop=\"A=B\"\tset plugin configuration property\n");
+#ifdef FLB_HAVE_PARSER
     printf("  -R, --parser=FILE\tspecify a parser configuration file\n");
+#endif
     printf("  -e, --plugin=FILE\tload an external plugin (shared lib)\n");
     printf("  -l, --log_file=FILE\twrite log info to a file\n");
     printf("  -t, --tag=TAG\t\tset plugin tag, same as '-p tag=abc'\n");
+#ifdef FLB_HAVE_STREAM_PROCESSOR
+    printf("  -T, --sp-task=SQL\tdefine a stream processor task\n");
+#endif
     printf("  -v, --verbose\t\tenable verbose mode\n");
 #ifdef FLB_HAVE_HTTP_SERVER
     printf("  -H, --http\t\tenable monitoring HTTP server\n");
     printf("  -P, --port\t\tset HTTP server TCP port (default: %s)\n",
            FLB_CONFIG_HTTP_PORT);
 #endif
+    printf("  -s, --coro_stack_size\tSet coroutines stack size in bytes "
+           "(default: %i)\n", config->coro_stack_size);
     printf("  -q, --quiet\t\tquiet mode\n");
     printf("  -S, --sosreport\tsupport report for Enterprise customers\n");
     printf("  -V, --version\t\tshow version number\n");
@@ -156,7 +163,7 @@ static void flb_help(int rc, struct flb_config *config)
     /* Iterate each supported input */
     mk_list_foreach(head, &config->in_plugins) {
         in = mk_list_entry(head, struct flb_input_plugin, _head);
-        if (strcmp(in->name, "lib") == 0) {
+        if (strcmp(in->name, "lib") == 0 || (in->flags & FLB_INPUT_PRIVATE)) {
             /* useless..., just skip it. */
             continue;
         }
@@ -192,8 +199,10 @@ static void flb_version()
 
 static void flb_banner()
 {
-    printf("%sFluent-Bit v%s%s\n", ANSI_BOLD, FLB_VERSION_STR, ANSI_RESET);
-    printf("%sCopyright (C) Treasure Data%s\n\n", ANSI_BOLD ANSI_YELLOW, ANSI_RESET);
+    fprintf(stderr, "%sFluent Bit v%s%s\n",
+            ANSI_BOLD, FLB_VERSION_STR, ANSI_RESET);
+    fprintf(stderr, "%sCopyright (C) Treasure Data%s\n\n",
+            ANSI_BOLD ANSI_YELLOW, ANSI_RESET);
 }
 
 #define flb_print_signal(X) case X:                       \
@@ -343,24 +352,25 @@ static void flb_service_conf_err(struct mk_rconf_section *section, char *key)
 
 static int flb_service_conf_path_set(struct flb_config *config, char *file)
 {
-    char *p;
     char *end;
-    char path[PATH_MAX + 1];
+    char *path;
 
-    p = realpath(file, path);
-    if (!p) {
+    path = realpath(file, NULL);
+    if (!path) {
         return -1;
     }
 
     /* lookup path ending and truncate */
-    end = strrchr(path, '/');
+    end = strrchr(path, FLB_DIRCHAR);
     if (!end) {
+        free(path);
         return -1;
     }
 
     end++;
     *end = '\0';
     config->conf_path = flb_strdup(path);
+    free(path);
 
     return 0;
 }
@@ -379,7 +389,12 @@ static int flb_service_conf(struct flb_config *config, char *file)
     struct flb_output_instance *out;
     struct flb_filter_instance *filter;
 
+#ifdef FLB_HAVE_STATIC_CONF
+    fconf = flb_config_static_open(file);
+#else
     fconf = mk_rconf_open(file);
+#endif
+
     if (!fconf) {
         return -1;
     }
@@ -451,7 +466,7 @@ static int flb_service_conf(struct flb_config *config, char *file)
 
         /* Create an instace of the plugin */
         tmp = flb_env_var_translate(config->env, name);
-        in = flb_input_new(config, tmp, NULL);
+        in = flb_input_new(config, tmp, NULL, FLB_TRUE);
         mk_mem_free(name);
         if (!in) {
             fprintf(stderr, "Input plugin '%s' cannot be loaded\n", tmp);
@@ -575,40 +590,43 @@ int main(int argc, char **argv)
     flb_stacktrace_init(argv[0]);
 #endif
 
-#ifndef _WIN32
     /* Setup long-options */
     static const struct option long_opts[] = {
-        { "buf_path",    required_argument, NULL, 'b' },
-        { "buf_workers", required_argument, NULL, 'B' },
-        { "config",      required_argument, NULL, 'c' },
+        { "storage_path",    required_argument, NULL, 'b' },
+        { "config",          required_argument, NULL, 'c' },
 #ifdef FLB_HAVE_FORK
-        { "daemon",      no_argument      , NULL, 'd' },
+        { "daemon",          no_argument      , NULL, 'd' },
 #endif
-        { "flush",       required_argument, NULL, 'f' },
-        { "http",        no_argument      , NULL, 'H' },
-        { "log_file",    required_argument, NULL, 'l' },
-        { "port",        required_argument, NULL, 'P' },
-        { "input",       required_argument, NULL, 'i' },
-        { "match",       required_argument, NULL, 'm' },
-        { "output",      required_argument, NULL, 'o' },
-        { "filter",      required_argument, NULL, 'F' },
-        { "parser",      required_argument, NULL, 'R' },
-        { "prop",        required_argument, NULL, 'p' },
-        { "plugin",      required_argument, NULL, 'e' },
-        { "tag",         required_argument, NULL, 't' },
-        { "version",     no_argument      , NULL, 'V' },
-        { "verbose",     no_argument      , NULL, 'v' },
-        { "quiet",       no_argument      , NULL, 'q' },
-        { "help",        no_argument      , NULL, 'h' },
-        { "sosreport",   no_argument      , NULL, 'S' },
+        { "flush",           required_argument, NULL, 'f' },
+        { "http",            no_argument      , NULL, 'H' },
+        { "log_file",        required_argument, NULL, 'l' },
+        { "port",            required_argument, NULL, 'P' },
+        { "input",           required_argument, NULL, 'i' },
+        { "match",           required_argument, NULL, 'm' },
+        { "output",          required_argument, NULL, 'o' },
+        { "filter",          required_argument, NULL, 'F' },
+#ifdef FLB_HAVE_PARSER
+        { "parser",          required_argument, NULL, 'R' },
+#endif
+        { "prop",            required_argument, NULL, 'p' },
+        { "plugin",          required_argument, NULL, 'e' },
+        { "tag",             required_argument, NULL, 't' },
+#ifdef FLB_HAVE_STREAM_PROCESSOR
+        { "sp-task",         required_argument, NULL, 'T' },
+#endif
+        { "version",         no_argument      , NULL, 'V' },
+        { "verbose",         no_argument      , NULL, 'v' },
+        { "quiet",           no_argument      , NULL, 'q' },
+        { "help",            no_argument      , NULL, 'h' },
+        { "coro_stack_size", required_argument, NULL, 's'},
+        { "sosreport",       no_argument      , NULL, 'S' },
 #ifdef FLB_HAVE_HTTP_SERVER
-        { "http_server", no_argument      , NULL, 'H' },
-        { "http_listen", required_argument, NULL, 'L' },
-        { "http_port",   required_argument, NULL, 'P' },
+        { "http_server",     no_argument      , NULL, 'H' },
+        { "http_listen",     required_argument, NULL, 'L' },
+        { "http_port",       required_argument, NULL, 'P' },
 #endif
         { NULL, 0, NULL, 0 }
     };
-#endif
 
 #ifdef FLB_HAVE_MTRACE
     /* Start tracing malloc and free */
@@ -640,22 +658,18 @@ int main(int argc, char **argv)
         exit(EXIT_FAILURE);
     }
 
+#ifndef FLB_HAVE_STATIC_CONF
+
     /* Parse the command line options */
-    while ((opt = getopt_long(argc, argv, "b:B:c:df:i:m:o:R:F:p:e:t:l:vqVhL:HP:S",
+    while ((opt = getopt_long(argc, argv,
+                              "b:c:df:i:m:o:R:F:p:e:"
+                              "t:T:l:vqVhL:HP:s:S",
                               long_opts, NULL)) != -1) {
 
         switch (opt) {
-#ifdef FLB_HAVE_BUFFERING
         case 'b':
-            config->buffer_path = flb_strdup(optarg);
-            if (config->buffer_workers <= 0) {
-                config->buffer_workers = 1;
-            }
+            config->storage_path = flb_strdup(optarg);
             break;
-        case 'B':
-            config->buffer_workers = atoi(optarg);
-            break;
-#endif
         case 'c':
             cfg_file = flb_strdup(optarg);
             break;
@@ -664,23 +678,17 @@ int main(int argc, char **argv)
             config->daemon = FLB_TRUE;
             break;
 #endif
-
-#ifdef FLB_HAVE_PROXY_GO
         case 'e':
-            if (!flb_plugin_proxy_create(optarg, 0, config)) {
+            ret = flb_plugin_load_router(optarg, config);
+            if (ret == -1) {
                 exit(EXIT_FAILURE);
             }
             break;
-#else
-        case 'e':
-            fprintf(stderr, "Error: proxy Golang plugin not available\n");
-            exit(EXIT_FAILURE);
-#endif
         case 'f':
-            config->flush = atoi(optarg);
+            config->flush = atof(optarg);
             break;
         case 'i':
-            in = flb_input_new(config, optarg, NULL);
+            in = flb_input_new(config, optarg, NULL, FLB_TRUE);
             if (!in) {
                 flb_utils_error(FLB_ERR_INPUT_INVALID);
             }
@@ -701,7 +709,7 @@ int main(int argc, char **argv)
             }
             last_plugin = PLUGIN_OUTPUT;
             break;
-#ifdef FLB_HAVE_REGEX
+#ifdef FLB_HAVE_PARSER
         case 'R':
             ret = flb_parser_conf_file(optarg, config);
             if (ret != 0) {
@@ -738,6 +746,11 @@ int main(int argc, char **argv)
                 flb_input_set_property(in, "tag", optarg);
             }
             break;
+#ifdef FLB_HAVE_STREAM_PROCESSOR
+        case 'T':
+            flb_slist_add(&config->stream_processor_tasks, optarg);
+            break;
+#endif
         case 'h':
             flb_help(EXIT_SUCCESS, config);
             break;
@@ -767,6 +780,9 @@ int main(int argc, char **argv)
         case 'q':
             config->verbose = FLB_LOG_OFF;
             break;
+        case 's':
+            config->coro_stack_size = (unsigned int) atoi(optarg);
+            break;
         case 'S':
             config->support_mode = FLB_TRUE;
             break;
@@ -774,12 +790,14 @@ int main(int argc, char **argv)
             flb_help(EXIT_FAILURE, config);
         }
     }
+#endif /* !FLB_HAVE_STATIC_CONF */
 
     if (config->verbose != FLB_LOG_OFF) {
         flb_banner();
     }
 
     /* Validate config file */
+#ifndef FLB_HAVE_STATIC_CONF
     if (cfg_file) {
         if (access(cfg_file, R_OK) != 0) {
             flb_utils_error(FLB_ERR_CFG_FILE);
@@ -792,9 +810,20 @@ int main(int argc, char **argv)
         }
         flb_free(cfg_file);
     }
+#else
+    ret = flb_service_conf(config, "fluent-bit.conf");
+    if (ret != 0) {
+        flb_utils_error(FLB_ERR_CFG_FILE_STOP);
+    }
+#endif
+
+    /* Check co-routine stack size */
+    if (config->coro_stack_size < getpagesize()) {
+        flb_utils_error(FLB_ERR_CORO_STACK_SIZE);
+    }
 
     /* Validate flush time (seconds) */
-    if (config->flush < 1) {
+    if (config->flush <= (double) 0.0) {
         flb_utils_error(FLB_ERR_CFG_FLUSH);
     }
 
