@@ -2,6 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
+ *  Copyright (C) 2019      The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -20,12 +21,15 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_http_server.h>
 
+#include "uptime.h"
 #include "metrics.h"
 #include "plugins.h"
 
 int api_v1_registration(struct flb_hs *hs)
 {
+    api_v1_uptime(hs);
     api_v1_metrics(hs);
     api_v1_plugins(hs);
+
     return 0;
 }
